import logging
from typing import Optional

import a_sync

from y import convert
from y.constants import CHAINID
from y.datatypes import AnyAddressType, Block, UsdPrice
from y.networks import Network
from y.utils.raw_calls import raw_call

logger = logging.getLogger(__name__)

POOL = "0x83E5f18Da720119fF363cF63417628eB0e9fd523"


def is_froyo(token: AnyAddressType) -> bool:
    """
    Check if a given token is the Froyo token on the Fantom network.

    Args:
        token: The token address to check.

    Returns:
        True if the token is the Froyo token on Fantom, False otherwise.

    Examples:
        >>> is_froyo("0x4f85Bbf3B0265DCEd4Ec72ebD0358ccCf190F1B3")
        True
        >>> is_froyo("0x0000000000000000000000000000000000000000")
        False

    See Also:
        - :func:`y.convert.to_address` for address conversion.
        - :class:`y.networks.Network` for network identification.
    """
    address = convert.to_address(token)
    return (
        CHAINID == Network.Fantom
        and address == "0x4f85Bbf3B0265DCEd4Ec72ebD0358ccCf190F1B3"
    )


@a_sync.a_sync(default="sync")
async def get_price(
    token: AnyAddressType, block: Optional[Block] = None
) -> Optional[UsdPrice]:
    """
    Get the USD price of the Froyo pool token on the Fantom network.

    This function retrieves the virtual price of the Froyo pool token if the provided
    token address matches the Froyo pool address on the Fantom network. The price is
    returned as a :class:`y.datatypes.UsdPrice` object.

    Args:
        token: The token address to get the price for.
        block: The block number to query the price at, defaults to the latest block.

    Returns:
        The USD price of the Froyo pool token, or None if the token is not the Froyo pool.

    Examples:
        >>> await get_price("0x83E5f18Da720119fF363cF63417628eB0e9fd523")
        UsdPrice(1.23456789)
        >>> await get_price("0x0000000000000000000000000000000000000000")
        None

    See Also:
        - :func:`y.utils.raw_calls.raw_call` for making raw contract calls.
        - :class:`y.datatypes.UsdPrice` for the price representation.
    """
    if token != POOL:
        return None
    virtual_price = await raw_call(
        POOL, "get_virtual_price()", block=block, output="int", sync=False
    )
    return UsdPrice(virtual_price / 10**18)



from functools import lru_cache

from y._db.utils.utils import ensure_chain

"""
This module provides utility functions to retrieve the `get_block` and `get_token` 
functions from either the `eth_portfolio` or `ypricemagic` database utilities, 
depending on the availability of the `eth_portfolio` module.

These functions are used to ensure that the correct version of the database 
utilities is used, allowing for extended functionality if `eth_portfolio` is installed.
"""


@lru_cache(maxsize=1)
def _get_get_block():
    """Retrieve the `get_block` function from the appropriate module.

    This function first ensures that the chain object for the connected chain
    has been inserted into the database. It then attempts to import the
    `get_block` function from `eth_portfolio._db.utils` if the `eth_portfolio`
    module is installed. If the module is not installed, it falls back to
    importing `get_block` from `y._db.utils`.

    Examples:
        >>> get_block = _get_get_block()
        >>> block = get_block(12345678)
        >>> print(block)

    See Also:
        - :func:`y._db.utils.utils.ensure_chain`
        - :func:`y._db.utils.utils.get_block`
    """
    ensure_chain()
    try:
        from eth_portfolio._db.utils import get_block
    except ModuleNotFoundError:
        from y._db.utils import get_block
    return get_block


@lru_cache(maxsize=1)
def _get_get_token():
    """Retrieve the `get_token` function from the appropriate module.

    This function first ensures that the chain object for the connected chain
    has been inserted into the database. It then attempts to import the
    `get_token` function from `eth_portfolio._db.utils` if the `eth_portfolio`
    module is installed. If the module is not installed, it falls back to
    importing `get_token` from `y._db.utils.token`.

    Additionally, it forces imports to run in the main thread to ensure
    compatibility.

    Examples:
        >>> get_token = _get_get_token()
        >>> token = get_token("0x1234567890abcdef1234567890abcdef12345678")
        >>> print(token)

    See Also:
        - :func:`y._db.utils.utils.ensure_chain`
        - :func:`y._db.utils.token.get_token`
    """
    ensure_chain()
    try:
        from eth_portfolio._db.utils import get_token
    except ModuleNotFoundError:
        from y._db.utils.token import get_token
    # force imports to run in main thread
    import y._db.utils.token

    return get_token



from typing import Optional

import a_sync
from eth_typing import ChecksumAddress

from y import ENVIRONMENT_VARIABLES as ENVS
from y.datatypes import Block, UsdPrice
from y.prices import magic

MAPPING = {
    "0x30D9410ED1D5DA1F6C8391af5338C93ab8d4035C": "0x6c3F90f043a72FA612cbac8115EE7e52BDe6E490",  # cvx3crv -> 3crv
    "0xbE0F6478E0E4894CFb14f32855603A083A57c7dA": "0xd632f22692FaC7611d2AA1C0D552930D43CAEd3B",  # cvxFRAX3CRV-f -> FRAX3CRV
    "0xabB54222c2b77158CC975a2b715a3d703c256F05": "0x5a6A4D54456819380173272A5E8E9B9904BdF41B",  # cvxMIM-3LP3CRV-f -> crvMIM
    "0xCA3D9F45FfA69ED454E66539298709cb2dB8cA61": "0x43b4FdFD4Ff969587185cDB6f0BD875c5Fc83f8c",  # cvxalUSD3CRV-f -> crvalusd
    "0xDeFd8FdD20e0f34115C7018CCfb655796F6B2168": "0xc4AD29ba4B3c580e6D59105FFf484999997675Ff",  # cvx3crypto -> 3crypto
}


def is_convex_lp(token_address: ChecksumAddress) -> bool:
    """
    Check if a given token address is a Convex LP token.

    Args:
        token_address: The address of the token to check.

    Returns:
        True if the token is a Convex LP token, False otherwise.

    Examples:
        >>> is_convex_lp("0x30D9410ED1D5DA1F6C8391af5338C93ab8d4035C")
        True

        >>> is_convex_lp("0x0000000000000000000000000000000000000000")
        False
    """
    return token_address in MAPPING


@a_sync.a_sync(default="sync")
async def get_price(
    token_address: ChecksumAddress,
    block: Optional[Block] = None,
    skip_cache: bool = ENVS.SKIP_CACHE,
) -> UsdPrice:
    """
    Get the price of a Convex LP token in USD.

    This function retrieves the price of the underlying token mapped to the given Convex LP token address.

    Args:
        token_address: The address of the Convex LP token.
        block: The block number at which to fetch the price. Defaults to the latest block.
        skip_cache: Whether to skip the cache when fetching the price. Defaults to the value of ENVS.SKIP_CACHE.

    Returns:
        The price of the token in USD.

    Examples:
        >>> await get_price("0x30D9410ED1D5DA1F6C8391af5338C93ab8d4035C")
        1.23456789

    See Also:
        - :func:`y.prices.magic.get_price` for more details on how the price is fetched.
    """
    return await magic.get_price(
        MAPPING[token_address], block, skip_cache=skip_cache, sync=False
    )



import logging
from decimal import Decimal
from typing import Optional

import a_sync
from a_sync import cgather

from y import ENVIRONMENT_VARIABLES as ENVS
from y import convert
from y.constants import CHAINID, CONNECTED_TO_MAINNET, weth
from y.datatypes import AnyAddressType, Block, UsdPrice
from y.networks import Network
from y.prices import magic
from y.utils.raw_calls import raw_call

logger = logging.getLogger(__name__)


class wstEth(a_sync.ASyncGenericBase):
    """
    A class to interact with the wstETH token on Ethereum Mainnet.

    This class provides methods to fetch the price of wstETH in USD, leveraging the `a_sync` library
    to support both synchronous and asynchronous operations.

    Examples:
        Initialize the class with asynchronous behavior:

        >>> wsteth = wstEth(asynchronous=True)
        >>> price = await wsteth.get_price(block=12345678)

        Initialize the class with synchronous behavior:

        >>> wsteth_sync = wstEth(asynchronous=False)
        >>> price_sync = wsteth_sync.get_price(block=12345678)

    See Also:
        - :class:`a_sync.ASyncGenericBase`
    """

    def __init__(self, *, asynchronous: bool = False) -> None:
        """
        Initialize the wstEth class with optional asynchronous behavior.

        Args:
            asynchronous: A boolean indicating if the operations should be asynchronous.

        Attributes:
            address: The address of the wstETH token on the Ethereum Mainnet, or None if not on Mainnet.
            wrapped_for_curve: The address of the wrapped version of wstETH for Curve on the Ethereum Mainnet, or None if not on Mainnet.
        """
        super().__init__()
        self.asynchronous = asynchronous
        try:
            self.address = {
                Network.Mainnet: "0x7f39C581F595B53c5cb19bD0b3f8dA6c935E2Ca0"
            }[CHAINID]
            self.wrapped_for_curve = {
                Network.Mainnet: "0xb82CFa4325568748506dC7cF267857Ff1e3b8d39"
            }[CHAINID]
        except KeyError:
            self.address = None

    async def get_price(
        self, block: Optional[Block] = None, skip_cache: bool = ENVS.SKIP_CACHE
    ) -> UsdPrice:
        """
        Fetch the price of wstETH in USD.

        This method retrieves the price of wstETH by calling the `stEthPerToken` method
        on the wstETH contract using :func:`y.utils.raw_calls.raw_call` and fetching the current price
        of WETH using :func:`y.prices.magic.get_price`.

        Args:
            block: The block number at which to fetch the price.
            skip_cache: Whether to skip the cache when fetching the price.

        Examples:
            Fetch the price asynchronously:

            >>> price = await wsteth.get_price(block=12345678)

            Fetch the price synchronously:

            >>> price_sync = wsteth_sync.get_price(block=12345678)

        See Also:
            - :func:`y.utils.raw_calls.raw_call`
            - :func:`y.prices.magic.get_price`
        """
        share_price, weth_price = await cgather(
            raw_call(
                self.address, "stEthPerToken()", output="int", block=block, sync=False
            ),
            magic.get_price(weth, block, skip_cache=skip_cache, sync=False),
        )
        share_price /= Decimal(10**18)
        return UsdPrice(share_price * Decimal(float(weth_price)))


wsteth = wstEth(asynchronous=True)

if CONNECTED_TO_MAINNET:
    _WSTETH_ADDRS = (wsteth.address, wsteth.wrapped_for_curve)


def is_wsteth(address: AnyAddressType) -> bool:
    """
    Check if a given address is wstETH or its wrapped version for Curve on Ethereum Mainnet.

    This function verifies the network ID to ensure it is operating on the Ethereum Mainnet
    before performing the address comparison.

    Args:
        address: The address to check.

    Examples:
        Check if an address is wstETH:

        >>> is_wsteth("0x7f39C581F595B53c5cb19bD0b3f8dA6c935E2Ca0")
        True

        Check if an address is not wstETH:

        >>> is_wsteth("0x0000000000000000000000000000000000000000")
        False

    See Also:
        - :class:`wstEth`
    """
    return CONNECTED_TO_MAINNET and convert.to_address(address) in _WSTETH_ADDRS



import logging
from decimal import Decimal
from typing import Optional

import dank_mids
from a_sync import a_sync, cgather
from brownie import ZERO_ADDRESS, chain

from y import ENVIRONMENT_VARIABLES as ENVS
from y import convert
from y.classes.common import ERC20
from y.constants import weth
from y.contracts import Contract
from y.datatypes import AnyAddressType, Block, UsdPrice
from y.prices import magic
from y.utils import gather_methods

logger = logging.getLogger(__name__)

if chain.id == 1:
    router = Contract("0xbAF9A5d4b0052359326A6CDAb54BABAa3a3A9643")
    gas_coin = weth
elif chain.id == 56:
    router = Contract("0xD41B24bbA51fAc0E4827b6F94C0D6DDeB183cD64")
    gas_coin = Contract("0xbb4CdB9CBd36B01bD1cBaEBF2De08d9173bc095c")  # wbnb
else:
    router = None
    gas_coin = None


@a_sync(default="sync", ram_cache_ttl=5 * 60)
async def is_mooniswap_pool(token: AnyAddressType) -> bool:
    """
    Check if the given token address is a Mooniswap pool.

    Args:
        token: The address of the token to check.

    Returns:
        True if the token is a Mooniswap pool, False otherwise.

    Examples:
        >>> is_mooniswap_pool("0x1234567890abcdef1234567890abcdef12345678")
        True

        >>> is_mooniswap_pool("0xabcdefabcdefabcdefabcdefabcdefabcdef")
        False
    """
    address = await convert.to_address_async(token)
    return False if router is None else await router.isPool.coroutine(address)


@a_sync(default="sync")
async def get_pool_price(
    token: AnyAddressType,
    block: Optional[Block] = None,
    skip_cache: bool = ENVS.SKIP_CACHE,
) -> UsdPrice:
    """
    Get the price of the given Mooniswap pool token.

    Args:
        token: The address of the pool token.
        block (optional): The block number to get the price at. Defaults to latest block.
        skip_cache (optional): Whether to skip the cache. Defaults to :obj:`ENVS.SKIP_CACHE`.

    Returns:
        The price of the pool token in USD as a :class:`~y.datatypes.UsdPrice`.

    Examples:
        >>> get_pool_price("0x1234567890abcdef1234567890abcdef12345678")
        UsdPrice('1.2345')

        >>> get_pool_price("0xabcdefabcdefabcdefabcdefabcdefabcdef", block=12345678)
        UsdPrice('0.9876')

    See Also:
        - :func:`y.prices.magic.get_price`
    """
    address = await convert.to_address_async(token)
    token0, token1 = await gather_methods(address, ("token0", "token1"))
    bal0, bal1, price0, price1, total_supply = await cgather(
        (
            dank_mids.eth.get_balance(address, block_identifier=block)
            if token0 == ZERO_ADDRESS
            else ERC20(token0, asynchronous=True).balance_of_readable(address, block)
        ),
        (
            dank_mids.eth.get_balance(address, block_identifier=block)
            if token1 == ZERO_ADDRESS
            else ERC20(token1, asynchronous=True).balance_of_readable(address, block)
        ),
        (
            magic.get_price(gas_coin, block, skip_cache=skip_cache, sync=False)
            if token0 == ZERO_ADDRESS
            else magic.get_price(token0, block, skip_cache=skip_cache, sync=False)
        ),
        (
            magic.get_price(gas_coin, block, skip_cache=skip_cache, sync=False)
            if token1 == ZERO_ADDRESS
            else magic.get_price(token1, block, skip_cache=skip_cache, sync=False)
        ),
        ERC20(address, asynchronous=True).total_supply_readable(block),
    )

    if token0 == ZERO_ADDRESS:
        bal0 /= 10**18
    elif token1 == ZERO_ADDRESS:
        bal1 /= 10**18

    totalVal = bal0 * float(price0) + bal1 * float(price1)
    price = totalVal / total_supply
    return Decimal(price)



import functools
import os
from inspect import iscoroutinefunction
from logging import DEBUG, getLogger

import a_sync
import eth_retry

from a_sync._typing import AnyFn, P, T
from a_sync.a_sync.function import ASyncDecorator
from brownie import chain
from joblib import Memory

from y import ENVIRONMENT_VARIABLES as ENVS


@eth_retry.auto_retry
def _memory():
    """
    Create and return a :class:`Memory` object for caching values for the currently connected blockchain.

    See Also:
        - :class:`joblib.Memory`
    """
    return Memory(f"cache/{chain.id}", verbose=0)


memory = _memory()

a_sync_ttl_cache: ASyncDecorator = a_sync.a_sync(ram_cache_ttl=ENVS.CACHE_TTL)


# NOTE: we have an optional disk cache that I made for debugging and decided to keep in for my own convenience.
# TODO: make a real disk cache

try:

    import toolcache

    """User has toolcache, this diskcache decorator will work."""

    logger = getLogger(__name__)
    cache_base_path = f"./cache/{chain.id}/"

    def optional_async_diskcache(fn: AnyFn[P, T]) -> AnyFn[P, T]:
        """
        Decorator to optionally apply disk caching to asynchronous functions.

        If the user has `toolcache` installed, this decorator will apply disk caching
        to the decorated asynchronous function. If `toolcache` is not installed, the
        function will be returned as is, without any caching applied.

        Args:
            fn: The function to be decorated. Must be asynchronous.

        Raises:
            NotImplementedError: If the function is synchronous.

        Examples:
            Using the decorator with an asynchronous function when `toolcache` is installed:

            >>> @optional_async_diskcache
            ... async def fetch_data():
            ...     return "data"

            This will cache the result of `fetch_data` in the specified cache directory.

            If `toolcache` is not installed, the function will run without caching:

            >>> async def fetch_data():
            ...     return "data"
            >>> fetch_data = optional_async_diskcache(fetch_data)

        See Also:
            - :mod:`toolcache`
        """
        if not iscoroutinefunction(fn):
            raise NotImplementedError(f"{fn} is sync")

        module = fn.__module__.replace(".", "/")
        cache_path_for_fn = cache_base_path + module + "/" + fn.__name__

        # Ensure the cache dir exists
        os.makedirs(cache_path_for_fn, exist_ok=True)

        if logger.isEnabledFor(DEBUG):

            @toolcache.cache("disk", cache_dir=cache_path_for_fn)
            @functools.wraps(fn)
            async def diskcache_wrap_with_debug_logs(*args, **kwargs) -> T:
                logger.debug(
                    "fetching %s(%s)", fn.__qualname__, ", ".join(map(str, args))
                )
                return await fn(*args, **kwargs)

            return diskcache_wrap_with_debug_logs

        else:

            @toolcache.cache("disk", cache_dir=cache_path_for_fn)
            @functools.wraps(fn)
            async def diskcache_wrap(*args, **kwargs) -> T:
                return await fn(*args, **kwargs)

            return diskcache_wrap

except ImportError:

    """User doesn't have toolcache, this decorator will just return the undecorated function."""

    def optional_async_diskcache(fn: AnyFn[P, T]) -> AnyFn[P, T]:
        """
        Decorator to optionally apply disk caching to asynchronous functions.

        If `toolcache` is not installed, this decorator will return the function
        as is, without any caching applied.

        Args:
            fn: The function to be decorated.

        Examples:
            Using the decorator with an asynchronous function:

            >>> async def fetch_data():
            ...     return "data"
            >>> fetch_data = optional_async_diskcache(fetch_data)

        See Also:
            - :mod:`toolcache`
        """
        return fn



WeightedPoolABI = [
    {
        "inputs": [
            {"internalType": "contract IVault", "name": "vault", "type": "address"},
            {"internalType": "string", "name": "name", "type": "string"},
            {"internalType": "string", "name": "symbol", "type": "string"},
            {
                "internalType": "contract IERC20[]",
                "name": "tokens",
                "type": "address[]",
            },
            {
                "internalType": "uint256[]",
                "name": "normalizedWeights",
                "type": "uint256[]",
            },
            {"internalType": "uint256", "name": "swapFeePercentage", "type": "uint256"},
            {
                "internalType": "uint256",
                "name": "pauseWindowDuration",
                "type": "uint256",
            },
            {
                "internalType": "uint256",
                "name": "bufferPeriodDuration",
                "type": "uint256",
            },
            {"internalType": "address", "name": "owner", "type": "address"},
        ],
        "stateMutability": "nonpayable",
        "type": "constructor",
    },
    {
        "anonymous": False,
        "inputs": [
            {
                "indexed": True,
                "internalType": "address",
                "name": "owner",
                "type": "address",
            },
            {
                "indexed": True,
                "internalType": "address",
                "name": "spender",
                "type": "address",
            },
            {
                "indexed": False,
                "internalType": "uint256",
                "name": "value",
                "type": "uint256",
            },
        ],
        "name": "Approval",
        "type": "event",
    },
    {
        "anonymous": False,
        "inputs": [
            {"indexed": False, "internalType": "bool", "name": "paused", "type": "bool"}
        ],
        "name": "PausedStateChanged",
        "type": "event",
    },
    {
        "anonymous": False,
        "inputs": [
            {
                "indexed": False,
                "internalType": "uint256",
                "name": "swapFeePercentage",
                "type": "uint256",
            }
        ],
        "name": "SwapFeePercentageChanged",
        "type": "event",
    },
    {
        "anonymous": False,
        "inputs": [
            {
                "indexed": True,
                "internalType": "address",
                "name": "from",
                "type": "address",
            },
            {
                "indexed": True,
                "internalType": "address",
                "name": "to",
                "type": "address",
            },
            {
                "indexed": False,
                "internalType": "uint256",
                "name": "value",
                "type": "uint256",
            },
        ],
        "name": "Transfer",
        "type": "event",
    },
    {
        "inputs": [],
        "name": "DOMAIN_SEPARATOR",
        "outputs": [{"internalType": "bytes32", "name": "", "type": "bytes32"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "address", "name": "owner", "type": "address"},
            {"internalType": "address", "name": "spender", "type": "address"},
        ],
        "name": "allowance",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "address", "name": "spender", "type": "address"},
            {"internalType": "uint256", "name": "amount", "type": "uint256"},
        ],
        "name": "approve",
        "outputs": [{"internalType": "bool", "name": "", "type": "bool"}],
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "inputs": [{"internalType": "address", "name": "account", "type": "address"}],
        "name": "balanceOf",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [],
        "name": "decimals",
        "outputs": [{"internalType": "uint8", "name": "", "type": "uint8"}],
        "stateMutability": "pure",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "address", "name": "spender", "type": "address"},
            {"internalType": "uint256", "name": "amount", "type": "uint256"},
        ],
        "name": "decreaseApproval",
        "outputs": [{"internalType": "bool", "name": "", "type": "bool"}],
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "inputs": [{"internalType": "bytes4", "name": "selector", "type": "bytes4"}],
        "name": "getActionId",
        "outputs": [{"internalType": "bytes32", "name": "", "type": "bytes32"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [],
        "name": "getAuthorizer",
        "outputs": [
            {"internalType": "contract IAuthorizer", "name": "", "type": "address"}
        ],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [],
        "name": "getInvariant",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [],
        "name": "getLastInvariant",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [],
        "name": "getNormalizedWeights",
        "outputs": [{"internalType": "uint256[]", "name": "", "type": "uint256[]"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [],
        "name": "getOwner",
        "outputs": [{"internalType": "address", "name": "", "type": "address"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [],
        "name": "getPausedState",
        "outputs": [
            {"internalType": "bool", "name": "paused", "type": "bool"},
            {
                "internalType": "uint256",
                "name": "pauseWindowEndTime",
                "type": "uint256",
            },
            {
                "internalType": "uint256",
                "name": "bufferPeriodEndTime",
                "type": "uint256",
            },
        ],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [],
        "name": "getPoolId",
        "outputs": [{"internalType": "bytes32", "name": "", "type": "bytes32"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [],
        "name": "getRate",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [],
        "name": "getSwapFeePercentage",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [],
        "name": "getVault",
        "outputs": [{"internalType": "contract IVault", "name": "", "type": "address"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "address", "name": "spender", "type": "address"},
            {"internalType": "uint256", "name": "amount", "type": "uint256"},
        ],
        "name": "increaseApproval",
        "outputs": [{"internalType": "bool", "name": "", "type": "bool"}],
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "inputs": [],
        "name": "name",
        "outputs": [{"internalType": "string", "name": "", "type": "string"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [{"internalType": "address", "name": "owner", "type": "address"}],
        "name": "nonces",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "bytes32", "name": "poolId", "type": "bytes32"},
            {"internalType": "address", "name": "sender", "type": "address"},
            {"internalType": "address", "name": "recipient", "type": "address"},
            {"internalType": "uint256[]", "name": "balances", "type": "uint256[]"},
            {"internalType": "uint256", "name": "lastChangeBlock", "type": "uint256"},
            {
                "internalType": "uint256",
                "name": "protocolSwapFeePercentage",
                "type": "uint256",
            },
            {"internalType": "bytes", "name": "userData", "type": "bytes"},
        ],
        "name": "onExitPool",
        "outputs": [
            {"internalType": "uint256[]", "name": "", "type": "uint256[]"},
            {"internalType": "uint256[]", "name": "", "type": "uint256[]"},
        ],
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "bytes32", "name": "poolId", "type": "bytes32"},
            {"internalType": "address", "name": "sender", "type": "address"},
            {"internalType": "address", "name": "recipient", "type": "address"},
            {"internalType": "uint256[]", "name": "balances", "type": "uint256[]"},
            {"internalType": "uint256", "name": "lastChangeBlock", "type": "uint256"},
            {
                "internalType": "uint256",
                "name": "protocolSwapFeePercentage",
                "type": "uint256",
            },
            {"internalType": "bytes", "name": "userData", "type": "bytes"},
        ],
        "name": "onJoinPool",
        "outputs": [
            {"internalType": "uint256[]", "name": "", "type": "uint256[]"},
            {"internalType": "uint256[]", "name": "", "type": "uint256[]"},
        ],
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "inputs": [
            {
                "components": [
                    {
                        "internalType": "enum IVault.SwapKind",
                        "name": "kind",
                        "type": "uint8",
                    },
                    {
                        "internalType": "contract IERC20",
                        "name": "tokenIn",
                        "type": "address",
                    },
                    {
                        "internalType": "contract IERC20",
                        "name": "tokenOut",
                        "type": "address",
                    },
                    {"internalType": "uint256", "name": "amount", "type": "uint256"},
                    {"internalType": "bytes32", "name": "poolId", "type": "bytes32"},
                    {
                        "internalType": "uint256",
                        "name": "lastChangeBlock",
                        "type": "uint256",
                    },
                    {"internalType": "address", "name": "from", "type": "address"},
                    {"internalType": "address", "name": "to", "type": "address"},
                    {"internalType": "bytes", "name": "userData", "type": "bytes"},
                ],
                "internalType": "struct IPoolSwapStructs.SwapRequest",
                "name": "request",
                "type": "tuple",
            },
            {"internalType": "uint256", "name": "balanceTokenIn", "type": "uint256"},
            {"internalType": "uint256", "name": "balanceTokenOut", "type": "uint256"},
        ],
        "name": "onSwap",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "address", "name": "owner", "type": "address"},
            {"internalType": "address", "name": "spender", "type": "address"},
            {"internalType": "uint256", "name": "value", "type": "uint256"},
            {"internalType": "uint256", "name": "deadline", "type": "uint256"},
            {"internalType": "uint8", "name": "v", "type": "uint8"},
            {"internalType": "bytes32", "name": "r", "type": "bytes32"},
            {"internalType": "bytes32", "name": "s", "type": "bytes32"},
        ],
        "name": "permit",
        "outputs": [],
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "bytes32", "name": "poolId", "type": "bytes32"},
            {"internalType": "address", "name": "sender", "type": "address"},
            {"internalType": "address", "name": "recipient", "type": "address"},
            {"internalType": "uint256[]", "name": "balances", "type": "uint256[]"},
            {"internalType": "uint256", "name": "lastChangeBlock", "type": "uint256"},
            {
                "internalType": "uint256",
                "name": "protocolSwapFeePercentage",
                "type": "uint256",
            },
            {"internalType": "bytes", "name": "userData", "type": "bytes"},
        ],
        "name": "queryExit",
        "outputs": [
            {"internalType": "uint256", "name": "bptIn", "type": "uint256"},
            {"internalType": "uint256[]", "name": "amountsOut", "type": "uint256[]"},
        ],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "bytes32", "name": "poolId", "type": "bytes32"},
            {"internalType": "address", "name": "sender", "type": "address"},
            {"internalType": "address", "name": "recipient", "type": "address"},
            {"internalType": "uint256[]", "name": "balances", "type": "uint256[]"},
            {"internalType": "uint256", "name": "lastChangeBlock", "type": "uint256"},
            {
                "internalType": "uint256",
                "name": "protocolSwapFeePercentage",
                "type": "uint256",
            },
            {"internalType": "bytes", "name": "userData", "type": "bytes"},
        ],
        "name": "queryJoin",
        "outputs": [
            {"internalType": "uint256", "name": "bptOut", "type": "uint256"},
            {"internalType": "uint256[]", "name": "amountsIn", "type": "uint256[]"},
        ],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [{"internalType": "bool", "name": "paused", "type": "bool"}],
        "name": "setPaused",
        "outputs": [],
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "uint256", "name": "swapFeePercentage", "type": "uint256"}
        ],
        "name": "setSwapFeePercentage",
        "outputs": [],
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "inputs": [],
        "name": "symbol",
        "outputs": [{"internalType": "string", "name": "", "type": "string"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [],
        "name": "totalSupply",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "address", "name": "recipient", "type": "address"},
            {"internalType": "uint256", "name": "amount", "type": "uint256"},
        ],
        "name": "transfer",
        "outputs": [{"internalType": "bool", "name": "", "type": "bool"}],
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "address", "name": "sender", "type": "address"},
            {"internalType": "address", "name": "recipient", "type": "address"},
            {"internalType": "uint256", "name": "amount", "type": "uint256"},
        ],
        "name": "transferFrom",
        "outputs": [{"internalType": "bool", "name": "", "type": "bool"}],
        "stateMutability": "nonpayable",
        "type": "function",
    },
]



import abc
from typing import Generic, Optional, Tuple, Type, TypeVar

import a_sync

import y.ENVIRONMENT_VARIABLES as ENVS
from y import contracts
from y.classes._abc import LiquidityPool
from y._decorators import stuck_coro_debugger
from y.datatypes import AddressOrContract, AnyAddressType, Block, UsdPrice


class BalancerPool(LiquidityPool):
    """
    Represents a liquidity pool specific to the Balancer protocol.

    This class is a subclass of :class:`~y.classes._abc.LiquidityPool` and is intended to represent
    a Balancer pool. However, it currently does not implement any specific functionality or methods
    related to the Balancer protocol.

    Examples:
        >>> class MyBalancerPool(BalancerPool):
        ...     pass
        >>> pool = MyBalancerPool("0xAddress")
        >>> print(pool)
        <MyBalancerPool object at 0x...>

    See Also:
        - :class:`~y.classes._abc.LiquidityPool`
    """


_B = TypeVar("_B", bound=BalancerPool)


class BalancerABC(a_sync.ASyncGenericBase, Generic[_B]):
    """
    Abstract base class for interacting with Balancer pools.

    This class provides asynchronous methods to check if a token is a Balancer pool and to get the price
    of a pool. It uses the :mod:`a_sync` library for asynchronous operations.

    Examples:
        >>> class MyBalancer(BalancerABC):
        ...     _pool_type = MyBalancerPool
        ...     _check_methods = ("method1", "method2")
        ...     async def get_token_price(self, token_address, block=None, skip_cache=False):
        ...         return UsdPrice(100.0)
        >>> balancer = MyBalancer()
        >>> is_pool = await balancer.is_pool("0xTokenAddress")
        >>> print(is_pool)
        True

    See Also:
        - :class:`~y.classes._abc.LiquidityPool`
    """

    def __repr__(self) -> str:
        return f"<{type(self).__name__} object at {hex(id(self))}>"

    @a_sync.a_sync(ram_cache_ttl=5 * 60)
    @stuck_coro_debugger
    async def is_pool(self, token_address: AnyAddressType) -> bool:
        """
        Check if the given token address is a Balancer pool.

        Args:
            token_address: The address of the token to check.

        Returns:
            True if the token is a Balancer pool, False otherwise.

        Examples:
            >>> is_pool = await balancer.is_pool("0xTokenAddress")
            >>> print(is_pool)
            True
        """
        return await contracts.has_methods(
            token_address, self._check_methods, sync=False
        )

    @stuck_coro_debugger
    async def get_pool_price(
        self,
        pool_address: AnyAddressType,
        block: Optional[Block] = None,
        skip_cache: bool = ENVS.SKIP_CACHE,
    ) -> UsdPrice:
        """
        Get the price of a Balancer pool.

        Args:
            pool_address: The address of the pool.
            block: The block number at which to get the price. If None, uses the latest block.
            skip_cache: If True, bypasses ypricemagic's local caching mechanisms and forces a fresh calculation.

        Returns:
            The price of the pool as a :class:`~y.datatypes.UsdPrice` object.

        Examples:
            >>> price = await balancer.get_pool_price("0xPoolAddress")
            >>> print(price)
            100.0
        """
        return await self._pool_type(pool_address, asynchronous=True).get_pool_price(
            block=block, skip_cache=skip_cache
        )

    @property
    @abc.abstractmethod
    def _pool_type(self) -> Type[_B]:
        """
        The type of Balancer pool.

        This property should be overridden by subclasses to specify the type of Balancer pool.

        Examples:
            >>> class MyBalancer(BalancerABC):
            ...     _pool_type = MyBalancerPool
        """

    @property
    @abc.abstractmethod
    def _check_methods(self) -> Tuple[str]:
        """
        The methods to check for identifying a Balancer pool.

        This property should be overridden by subclasses to specify the methods used to identify a Balancer pool.

        Examples:
            >>> class MyBalancer(BalancerABC):
            ...     _check_methods = ("method1", "method2")
        """

    @abc.abstractmethod
    async def get_token_price(
        self,
        token_address: AddressOrContract,
        block: Optional[Block] = None,
        skip_cache: bool = ENVS.SKIP_CACHE,
    ) -> Optional[UsdPrice]:
        """
        Get the price of a token in a Balancer pool.

        This method should be implemented by subclasses to provide the logic for retrieving the token price.

        Args:
            token_address: The address of the token.
            block: The block number at which to get the price. If None, uses the latest block.
            skip_cache: If True, bypasses ypricemagic's local caching mechanisms and forces a fresh calculation.

        Examples:
            >>> price = await balancer.get_token_price("0xTokenAddress")
            >>> print(price)
            100.0
        """



import time

import pytest
from brownie import ZERO_ADDRESS, chain

from tests.fixtures import mainnet_only
from y.contracts import contract_creation_block_async
from y.networks import Network
from y.prices.chainlink import FEEDS, chainlink

feeds = set(FEEDS.keys())
feeds.update(
    {
        # Add feeds that *should* come from registry
        Network.Mainnet: [
            "0x0000000000000000000000000000000000000024",
            "0x000000000000000000000000000000000000007c",
            "0x000000000000000000000000000000000000009c",
            "0x0000000000000000000000000000000000000164",
            "0x0000000000000000000000000000000000000188",
            "0x000000000000000000000000000000000000019a",
            "0x000000000000000000000000000000000000022A",
            "0x0000000000000000000000000000000000000236",
            "0x0000000000000000000000000000000000000260",
            # "0x0000000000000000000000000000000000000283", chainlink feed disabled at block 14673916 - 0xd1898776f4f2333c26c170b58ec98430d29c224a81256fd70add5db05a036b20
            "0x00000000000000000000000000000000000002be",
            "0x00000000000000000000000000000000000002c6",
            "0x00000000000000000000000000000000000002F4",
            "0x000000000000000000000000000000000000033a",
            "0x00000000000000000000000000000000000003d2",
            "0x00000000000000000000000000000000000003Da",
            "0x0bc529c00C6401aEF6D220BE8C6Ea1667F6Ad93e",
            "0x0D8775F648430679A709E98d2b0Cb6250d2887EF",
            "0x111111111117dC0aa78b770fA6A738034120C302",
            "0x1776e1F26f98b1A5dF9cD347953a26dd3Cb46671",
            "0x1F573D6Fb3F13d689FF844B4cE37794d79a7FF1C",
            "0x1f9840a85d5aF5bf1D1762F925BDADdC4201F984",
            "0x221657776846890989a759BA2973e427DfF5C9bB",
            "0x2260FAC5E5542a773Aa44fBCfeDf7C193bc2C599",
            "0x3155BA85D5F96b2d030a4966AF206230e46849cb",
            "0x33D0568941C0C64ff7e0FB4fbA0B11BD37deEd9f",
            "0x3432B6A60D23Ca0dFCa7761B7ab56459D9C964D0",
            "0x3472A5A71965499acd81997a54BBA8D852C6E53d",
            "0x3845badAde8e6dFF049820680d1F14bD3903a5d0",
            "0x408e41876cCCDC0F92210600ef50372656052a38",
            "0x4575f41308EC1483f3d399aa9a2826d74Da13Deb",
            "0x4688a8b1F292FDaB17E9a90c8Bc379dC1DBd8713",
            "0x4C19596f5aAfF459fA38B0f7eD92F11AE6543784",
            "0x4Fabb145d64652a948d72533023f6E7A623C7C53",
            "0x514910771AF9Ca656af840dff83E8264EcF986CA",
            "0x57Ab1ec28D129707052df4dF418D58a2D46d5f51",
            "0x584bC13c7D411c00c01A62e8019472dE68768430",
            "0x674C6Ad92Fd080e4004b2312b45f796a192D27a0",
            "0x6B175474E89094C44Da98b954EedeAC495271d0F",
            "0x6B3595068778DD592e39A122f4f5a5cF09C90fE2",
            "0x6f259637dcD74C767781E37Bc6133cd6A68aa161",
            "0x75231F58b43240C9718Dd58B4967c5114342a86c",
            "0x7D1AfA7B718fb893dB30A3aBc0Cfc608AaCfeBB0",
            "0x7Fc66500c84A76Ad7e9c93437bFc5Ac33E2DDaE9",
            "0x8290333ceF9e6D528dD5618Fb97a76f268f3EDD4",
            "0x84cA8bc7997272c7CfB4D0Cd3D55cd942B3c9419",
            "0x8Ab7404063Ec4DBcfd4598215992DC3F8EC853d7",
            "0x8CE9137d39326AD0cD6491fb5CC0CbA0e089b6A9",
            "0x967da4048cD07aB37855c090aAF366e4ce1b9F48",
            "0x9f8F72aA9304c8B593d555F12eF6589cC3A579A2",
            "0xA0b73E1Ff0B80914AB6fe0444E65848C4C34450b",
            "0xA0b86991c6218b36c1d19D4a2e9Eb0cE3606eB48",
            "0xa3BeD4E1c75D00fa6f4E5E6922DB7261B5E9AcD2",
            "0xA9B1Eb5908CfC3cdf91F9B8B3a74108598009096",
            "0xADE00C28244d5CE17D72E40330B1c318cD12B7c3",
            "0xBA11D00c5f74255f56a5E366F4F77f5A186d7f55",
            "0xbBbBBBBbbBBBbbbBbbBbbbbBBbBbbbbBbBbbBBbB",
            "0xBBbbCA6A901c926F240b89EacB641d8Aec7AEafD",
            "0xbE9375C6a420D2eEB258962efB95551A5b722803",
            "0xc00e94Cb662C3520282E6f5717214004A7f26888",
            "0xC011a73ee8576Fb46F5E1c5751cA3B9Fe0af2a6F",
            "0xC02aaA39b223FE8D0A0e5C4F27eAD9083C756Cc2",
            "0xc12eCeE46ed65D970EE5C899FCC7AE133AfF9b03",
            "0xC581b735A1688071A1746c968e0798D642EDE491",
            "0xcB3df3108635932D912632ef7132d03EcFC39080",
            "0xd26114cd6EE289AccF82350c8d8487fedB8A0C07",
            "0xD46bA6D942050d489DBd938a2C909A5d5039A161",
            "0xD533a949740bb3306d119CC777fa900bA034cd52",
            "0xD71eCFF9342A5Ced620049e616c5035F1dB98620",
            "0xdAC17F958D2ee523a2206206994597C13D831ec7",
            "0xdB25f211AB05b1c97D595516F45794528a807ad8",
            "0xdd974D5C2e2928deA5F71b9825b8b646686BD200",
            "0xe1aFe1Fd76Fd88f78cBf599ea1846231B8bA3B6B",
            "0xe28b3B32B6c345A34Ff64674606124Dd5Aceca30",
            "0xE41d2489571d322189246DaFA5ebDe1F4699F498",
            "0xEeeeeEeeeEeEeeEeEeEeeEEEeeeeEeeeeeeeEEeE",
            "0xf8C3527CC04340b208C854E985240c02F7B7793f",
            "0xfF20817765cB7f73d4bde2e66e067E58D11095C2",
        ],
        Network.Fantom: [
            "0x21be370D5312f44cB42ce377BC9b8a0cEF1A4C83",
            "0x321162Cd933E2Be498Cd2267a90534A804051b11",
            "0x2406dCe4dA5aB125A18295f4fB9FD36a0f7879A2",
            "0x74b23882a30290451A17c44f4F05243b6b58C76d",
            "0xBDC8fd437C489Ca3c6DA3B5a336D11532a532303",
            "0xd6070ae98b8069de6B494332d1A1a81B6179D960",
            "0x1E4F97b9f9F913c46F1632781732927B9019C68b",
            "0x6a07A792ab2965C72a5B8088d3a069A7aC3a993B",
            "0x657A1861c15A3deD9AF0B6799a195a249ebdCbc6",
            "0xb3654dc3D10Ea7645f8319668E8F54d2574FBdC8",
            "0x399fe752D39338d28C36F3370fbebd8292fb9E6e",
            "0x56ee926bD8c72B2d5fa1aF4d9E4Cbb515a1E3Adc",
            "0x468003B688943977e6130F4F68F23aad939a1040",
            "0xae75A438b2E0cB8Bb01Ec1E1e376De11D44477CC",
            "0x81740D647493a61329E1c574A11ee7577659fb14",
            "0xe105621721D1293c27be7718e041a4Ce0EbB227E",
            "0x29b0Da86e484E1C0029B56e817912d778aC0EC69",
        ],
    }.get(chain.id, [])
)
FEEDS = list(feeds)


@pytest.mark.parametrize("token", FEEDS)
@pytest.mark.asyncio_cooperative
async def test_chainlink_get_feed(token):
    """
    Test `chainlink.get_feed` with both lowercase address and checksum address.

    This test ensures that the `chainlink.get_feed` function returns a valid feed
    address that is not the zero address for a given token.

    Args:
        token: The token address to test.

    Raises:
        AssertionError: If no feed is available for the token.
    """
    assert (
        await chainlink.get_feed(token, sync=False) != ZERO_ADDRESS
    ), "no feed available"


@pytest.mark.parametrize("token", FEEDS)
@pytest.mark.asyncio_cooperative
async def test_chainlink_latest(token):
    """
    Test the latest price availability for a token using Chainlink.

    This test checks if the latest price for a given token is available using
    Chainlink. If the price is not available, it verifies the feed's latest
    timestamp and skips the test if the feed is stale. Otherwise, it asserts
    that the feed's contract aggregator is the zero address, indicating no
    current price is available.

    Args:
        token: The token address to test.

    Raises:
        AttributeError: If there is an issue accessing the feed's contract
            aggregator.
    """
    if not await chainlink.get_price(token):
        feed = await chainlink.get_feed(token)
        latest_timestamp = await feed.latest_timestamp()
        if latest_timestamp and latest_timestamp + 24 * 60 * 60 < time.time():
            pytest.skip("feed is stale")
        try:
            assert (
                await feed.contract.aggregator == ZERO_ADDRESS
            ), "no current price available"
        except AttributeError as e:
            raise AttributeError(*e.args, feed) from e


@mainnet_only
@pytest.mark.parametrize("token", FEEDS)
@pytest.mark.asyncio_cooperative
async def test_chainlink_before_registry(token):
    """
    Test Chainlink price retrieval before the registry was available.

    This test checks if a price can be retrieved for a given token at a block
    before the Chainlink registry was available. If the feed was deployed after
    the test block, the test is skipped. If the price is not available, it
    verifies the feed's latest timestamp and skips the test if the feed is
    stale. Otherwise, it asserts that the feed's contract aggregator is the
    zero address, indicating no price is available before the registry.

    Args:
        token: The token address to test.

    Raises:
        AttributeError: If there is an issue accessing the feed's contract
            aggregator.
    """
    test_block = 12800000
    assert chainlink.asynchronous is True
    feed = await chainlink.get_feed(token, sync=False)
    if await contract_creation_block_async(feed.address) > test_block:
        pytest.skip("not applicable to feeds deployed after test block")
    price = chainlink.get_price(token, block=test_block)
    price = await price
    if not price:
        feed = await chainlink.get_feed(token)
        latest_timestamp = await feed.latest_timestamp()
        if latest_timestamp and latest_timestamp + 24 * 60 * 60 < time.time():
            pytest.skip("feed is stale")
        try:
            assert (
                await feed.contract.aggregator == ZERO_ADDRESS
            ), f"{feed} no price available before registry"
        except AttributeError as e:
            raise AttributeError(*e.args, feed) from e


@pytest.mark.asyncio_cooperative
async def test_chainlink_nonexistent():
    """
    Test Chainlink feed and price retrieval for a nonexistent token.

    This test checks that attempting to retrieve a feed or price for a
    nonexistent token (zero address) returns None.
    """
    assert await chainlink.get_feed(ZERO_ADDRESS, sync=False) is None
    assert await chainlink.get_price(ZERO_ADDRESS, sync=False) is None


@mainnet_only
@pytest.mark.asyncio_cooperative
async def test_chainlink_before_feed():
    """
    Test Chainlink price retrieval one block before a feed is deployed.

    This test checks that attempting to retrieve a price for a token one block
    before its Chainlink feed is deployed returns None.
    """
    # try to fetch yfi price one block before feed is deployed
    assert (
        await chainlink.get_price(
            "0x0bc529c00C6401aEF6D220BE8C6Ea1667F6Ad93e", 12742718, sync=False
        )
        is None
    )



import logging
from typing import Optional

from a_sync import a_sync, cgather

from y import ENVIRONMENT_VARIABLES as ENVS
from y import convert
from y.classes.common import ERC20
from y.constants import CONNECTED_TO_MAINNET, weth
from y.datatypes import AnyAddressType, Block, UsdPrice
from y.prices import magic
from y.utils.raw_calls import raw_call

logger = logging.getLogger(__name__)


def is_creth(token: AnyAddressType) -> bool:
    """Check if a token is crETH.

    Args:
        token: The token address to check.

    Returns:
        True if the token is crETH on the Ethereum Mainnet, False otherwise.

    Example:
        >>> is_creth("0xcBc1065255cBc3aB41a6868c22d1f1C573AB89fd")
        True
    """
    return (
        CONNECTED_TO_MAINNET
        and convert.to_address(token) == "0xcBc1065255cBc3aB41a6868c22d1f1C573AB89fd"
    )


@a_sync(default="sync")
async def get_price_creth(
    token: AnyAddressType,
    block: Optional[Block] = None,
    skip_cache: bool = ENVS.SKIP_CACHE,
) -> UsdPrice:
    """Get the price of crETH in USD.

    This function retrieves the accumulated balance and total supply of crETH,
    divides the total balance by the total supply to calculate the value per token,
    and then multiplies this value by the current WETH price to determine the price
    of crETH in USD.

    Args:
        token: The crETH token address.
        block: The block number to query. Defaults to the latest block.
        skip_cache: If True, skip using the cache while fetching price data.

    Examples:
        >>> await get_price_creth("0xcBc1065255cBc3aB41a6868c22d1f1C573AB89fd")
        1234.56

    See Also:
        - :func:`y.prices.magic.get_price`
        - :class:`y.classes.common.ERC20`
    """
    address = await convert.to_address_async(token)
    total_balance, total_supply, weth_price = await cgather(
        raw_call(address, "accumulated()", output="int", block=block, sync=False),
        ERC20(address, asynchronous=True).total_supply(block),
        magic.get_price(weth, block, skip_cache=skip_cache, sync=False),
    )
    per_share = total_balance / total_supply
    return per_share * weth_price



from typing import Optional

import a_sync
from a_sync import cgather
from a_sync.a_sync import HiddenMethodDescriptor
from brownie.exceptions import VirtualMachineError
from typing_extensions import Self

from y import Contract
from y.classes.common import ERC20
from y.constants import CHAINID
from y.datatypes import Address, AddressOrContract, Block
from y.exceptions import UnsupportedNetwork
from y.networks import Network

addresses = {
    # https://docs.fantom.foundation/tutorials/band-protocol-standard-dataset
    Network.Fantom: "0x56E2898E0ceFF0D1222827759B56B28Ad812f92F"
}

supported_assets = {
    # https://docs.fantom.foundation/tutorials/band-protocol-standard-dataset#supported-tokens
    Network.Fantom: [
        "0xaf319E5789945197e365E7f7fbFc56B130523B33",  # FRAX
        "0x04068DA6C83AFCFA0e13ba15A6696662335D5B75",  # USDC
        "0x0E1694483eBB3b74d3054E383840C6cf011e518e",  # sUSD
        "0x6a07A792ab2965C72a5B8088d3a069A7aC3a993B",  # AAVE
        "0x46E7628E8b4350b2716ab470eE0bA1fa9e76c6C5",  # BAND
        "0xD67de0e0a0Fd7b15dC8348Bb9BE742F3c5850454",  # BNB
        "0x321162Cd933E2Be498Cd2267a90534A804051b11",  # BTC
        "0xB01E8419d842beebf1b70A7b5f7142abbaf7159D",  # COVER
        "0x657A1861c15A3deD9AF0B6799a195a249ebdCbc6",  # CREAM
        "0x1E4F97b9f9F913c46F1632781732927B9019C68b",  # CRV
        "0x74b23882a30290451A17c44f4F05243b6b58C76d",  # ETH
        "0x44B26E839eB3572c5E959F994804A5De66600349",  # HEGIC
        "0x2A5062D22adCFaAfbd5C541d4dA82E4B450d4212",  # KP3R
        "0xb3654dc3D10Ea7645f8319668E8F54d2574FBdC8",  # LINK
        "0x924828a9Fb17d47D0eb64b57271D10706699Ff11",  # SFI
        "0x56ee926bD8c72B2d5fa1aF4d9E4Cbb515a1E3Adc",  # SNX
        "0xae75A438b2E0cB8Bb01Ec1E1e376De11D44477CC",  # SUSHI
        "0x29b0Da86e484E1C0029B56e817912d778aC0EC69",  # YFI
    ]
}


class Band(a_sync.ASyncGenericSingleton):
    """A class to interact with the Band Protocol oracle on supported networks.

    This class provides methods to check asset support and fetch prices using the Band Protocol oracle.

    Raises:
        UnsupportedNetwork: If the Band Protocol is not supported on the current network.

    Examples:
        >>> band = Band(asynchronous=True)
        >>> "0xaf319E5789945197e365E7f7fbFc56B130523B33" in band
        True

    See Also:
        - :class:`a_sync.ASyncGenericSingleton`
    """

    def __init__(self, *, asynchronous: bool = False) -> None:
        """
        Initialize the Band class.

        Args:
            asynchronous: Whether to use asynchronous operations.

        Raises:
            UnsupportedNetwork: If the Band Protocol is not supported on the current network.

        Examples:
            >>> band = Band(asynchronous=True)
        """
        super().__init__()
        if CHAINID not in addresses:
            raise UnsupportedNetwork("band is not supported on this network")
        self.asynchronous = asynchronous
        super().__init__()

    def __contains__(self, asset: AddressOrContract) -> bool:
        """
        Check if an asset is supported by the Band Protocol on the current network.

        Args:
            asset: The asset address or contract to check.

        Examples:
            >>> band = Band(asynchronous=True)
            >>> "0xaf319E5789945197e365E7f7fbFc56B130523B33" in band
            True
        """
        return CHAINID in addresses and asset in supported_assets[CHAINID]

    @a_sync.aka.property
    async def oracle(self) -> Contract:
        """
        Get the Band Protocol oracle contract for the current network.

        Examples:
            >>> band = Band(asynchronous=True)
            >>> oracle = await band.oracle
        """
        return await Contract.coroutine(addresses[CHAINID])

    __oracle__: HiddenMethodDescriptor[Self, Contract]

    async def get_price(
        self, asset: Address, block: Optional[Block] = None
    ) -> Optional[float]:
        """
        Get the price of an asset in terms of USDC using the Band Protocol oracle.

        Args:
            asset: The asset address to get the price for.
            block: The block number to query the price at. Defaults to the latest block.

        Returns:
            The price of the asset in terms of USDC, or None if the price cannot be fetched.

        Examples:
            >>> band = Band(asynchronous=True)
            >>> price = await band.get_price("0xaf319E5789945197e365E7f7fbFc56B130523B33")
            >>> print(price)
            1.0
        """
        oracle, asset_symbol = await cgather(
            self.__oracle__,
            ERC20(asset, asynchronous=True).symbol,
        )
        try:
            reference_data = await oracle.getReferenceData.coroutine(
                asset_symbol, "USDC", block_identifier=block
            )
            return reference_data[0] / 10**18
        except ValueError:
            return None
        except VirtualMachineError:
            return None


try:
    band = Band(asynchronous=True)
except UnsupportedNetwork:
    band = set()



from decimal import Decimal
from typing import Optional

from a_sync import a_sync
from dank_mids.exceptions import Revert
from multicall.call import Call

from y import ENVIRONMENT_VARIABLES as ENVS
from y.classes.common import ERC20, WeiBalance
from y.contracts import Contract, has_methods
from y.datatypes import Address, Block
from y.utils.cache import optional_async_diskcache

METHODS = "main()(address)", "issuanceAvailable()(uint)", "redemptionAvailable()(uint)"


@a_sync(default="sync")
@optional_async_diskcache
async def is_rtoken(token_address: Address) -> bool:
    """
    Check if the given token is a Reserve Protocol R-token.

    Args:
        token_address: The address of the token to check.

    Returns:
        True if the token is a Reserve Protocol R-token, False otherwise.

    Example:
        >>> address = "0xA0b86991c6218b36c1d19D4a2e9Eb0cE3606eB48"  # USDC (not an R-token)
        >>> is_r = is_rtoken(address)
        >>> print(is_r)
        False
    """
    return await has_methods(token_address, METHODS, sync=False)


@a_sync(default="sync")
async def get_price(
    token_address: Address,
    block: Optional[Block] = None,
    skip_cache: bool = ENVS.SKIP_CACHE,
) -> Decimal:
    """
    Get the price of a Reserve Protocol R-token in USD.

    Args:
        token_address: The address of the R-token.
        block (optional): The block number to query. Defaults to None (latest).
        skip_cache (optional): Whether to skip cache. Defaults to :obj:`ENVS.SKIP_CACHE`.

    Returns:
        The price of the R-token in USD.

    Raises:
        TypeError: If the token is not a valid R-token.
        Exception: If unable to calculate the price.

    Example:
        >>> address = "0xaCeeD87BD5754c3d714F3Bd43a9B7B0C9250ab0D"  # RSV token
        >>> price = await get_price(address, sync=False)
        >>> print(price)
        1.00
    """
    main = await Call(token_address, "main()(address)", block_id=block)
    if main is None:
        raise TypeError(main, token_address, await is_rtoken(token_address))
    basket_handler = await Contract.coroutine(
        await Call(main, "basketHandler()(address)", block_id=block)
    )
    try:
        low, high = await basket_handler.price.coroutine(block_identifier=block)
    except Revert:
        return None
    else:
        return Decimal(low + high) // 2 / 10**18

    tokens, *_ = await basket_handler.getPrimeBasket.coroutine(block_identifier=block)
    tokens = [ERC20(token, asynchronous=True) for token in tokens]
    balances = [
        WeiBalance(
            balance, token, block=block, skip_cache=skip_cache, asynchronous=True
        )
        for token, balance in zip(
            tokens, await basket_handler.quantity.map(tokens, block_identifier=block)
        )
    ]
    print(balances)
    values = WeiBalance.value_usd.map(balances).values()
    print(values)
    value = sum(values)
    print(value)
    supply = await ERC20(token_address, asynchronous=True).total_supply_readable(block)
    print(f"ts: {supply}")
    raise Exception(value / supply)



import logging
from decimal import Decimal
from typing import List, Optional

import a_sync
from a_sync import cgather
from brownie import ZERO_ADDRESS
from multicall import Call

from y import ENVIRONMENT_VARIABLES as ENVS
from y.classes.common import ERC20
from y.contracts import has_method
from y.datatypes import Address, AnyAddressType, Block, UsdPrice, UsdValue
from y.exceptions import call_reverted
from y.prices import magic
from y.utils.raw_calls import raw_call

logger = logging.getLogger(__name__)


@a_sync.a_sync(default="sync", cache_type="memory")
async def is_pie(token: AnyAddressType) -> bool:
    """
    Check if a given token is a PieDAO token.

    Args:
        token: The address of the token to check.

    Returns:
        True if the token is a PieDAO token, False otherwise.

    Example:
        >>> is_pie("0x1234567890abcdef1234567890abcdef12345678")
        True
    """
    return await has_method(token, "getCap()(uint)", sync=False)


@a_sync.a_sync(default="sync")
async def get_price(
    pie: AnyAddressType,
    block: Optional[Block] = None,
    skip_cache: bool = ENVS.SKIP_CACHE,
) -> UsdPrice:
    """
    Get the price of a PieDAO token in USD.

    Args:
        pie: The address of the PieDAO token.
        block: The block number to query the price at. Defaults to the latest block.
        skip_cache: Whether to skip the cache when fetching the price. Defaults to the value of ENVS.SKIP_CACHE.

    Example:
        >>> get_price("0x1234567890abcdef1234567890abcdef12345678")
        123.45

    See Also:
        - :func:`get_tvl`
        - :class:`ERC20`
    """
    tvl, total_supply = await cgather(
        get_tvl(pie, block, skip_cache=skip_cache),
        ERC20(pie, asynchronous=True).total_supply_readable(block),
    )
    return UsdPrice(tvl / total_supply)


async def get_tokens(
    pie_address: Address, block: Optional[Block] = None
) -> List[ERC20]:
    """
    Get the list of tokens in a PieDAO token.

    Args:
        pie_address: The address of the PieDAO token.
        block: The block number to query the tokens at. Defaults to the latest block.

    Returns:
        A list of :class:`ERC20` objects representing the tokens in the PieDAO token.

    Example:
        >>> get_tokens("0x1234567890abcdef1234567890abcdef12345678")
        [ERC20('0xTokenAddress1'), ERC20('0xTokenAddress2')]

    Note:
        This function retrieves token addresses using a multicall and then creates :class:`ERC20` instances from those addresses.
    """
    return [
        ERC20(t)
        for t in await Call(pie_address, "getTokens()(address[])", block_id=block)
    ]


async def get_bpool(pie_address: Address, block: Optional[Block] = None) -> Address:
    """
    Get the Balancer pool address for a PieDAO token.

    Args:
        pie_address: The address of the PieDAO token.
        block: The block number to query the pool at. Defaults to the latest block.

    Returns:
        The address of the Balancer pool, or the PieDAO address if no pool is found.

    Example:
        >>> get_bpool("0x1234567890abcdef1234567890abcdef12345678")
        '0xBpoolAddress'
    """
    try:
        bpool = await raw_call(
            pie_address, "getBPool()", output="address", block=block, sync=False
        )
        return bpool if bpool != ZERO_ADDRESS else pie_address
    except Exception as e:
        if not call_reverted(e):
            raise
        return pie_address


async def get_tvl(
    pie_address: Address,
    block: Optional[Block] = None,
    skip_cache: bool = ENVS.SKIP_CACHE,
) -> UsdValue:
    """
    Get the total value locked (TVL) in a PieDAO token in USD.

    Args:
        pie_address: The address of the PieDAO token.
        block: The block number to query the TVL at. Defaults to the latest block.
        skip_cache: Whether to skip the cache when fetching the TVL. Defaults to the value of ENVS.SKIP_CACHE.

    Example:
        >>> get_tvl("0x1234567890abcdef1234567890abcdef12345678")
        1000000.00

    See Also:
        - :func:`get_value`
        - :func:`get_bpool`
        - :func:`get_tokens`
    """
    tokens: List[ERC20]
    pool, tokens = await cgather(
        get_bpool(pie_address, block), get_tokens(pie_address, block)
    )
    return await a_sync.map(
        get_value, tokens, bpool=pool, block=block, skip_cache=skip_cache
    ).sum(pop=True, sync=False)


async def get_balance(
    bpool: Address, token: ERC20, block: Optional[Block] = None
) -> Decimal:
    """
    Get the balance of a token in a Balancer pool.

    Args:
        bpool: The address of the Balancer pool.
        token: The :class:`ERC20` token to query.
        block: The block number to query the balance at. Defaults to the latest block.

    Example:
        >>> get_balance("0xBpoolAddress", ERC20("0xTokenAddress"))
        1000.0
    """
    balance, scale = await cgather(
        Call(token.address, ("balanceOf(address)(uint)", bpool), block_id=block),
        token.__scale__,
    )
    return Decimal(balance) / scale


async def get_value(
    bpool: Address,
    token: ERC20,
    block: Optional[Block] = None,
    skip_cache: bool = ENVS.SKIP_CACHE,
) -> UsdValue:
    """
    Get the USD value of a token in a Balancer pool.

    Args:
        bpool: The address of the Balancer pool.
        token: The :class:`ERC20` token to query.
        block: The block number to query the value at. Defaults to the latest block.
        skip_cache: Whether to skip the cache when fetching the value. Defaults to the value of ENVS.SKIP_CACHE.

    Example:
        >>> get_value("0xBpoolAddress", ERC20("0xTokenAddress"))
        500.0

    Note:
        This function calculates the value by multiplying the token balance in the pool by its price.

    See Also:
        - :func:`get_balance`
    """
    balance, price = await cgather(
        get_balance(bpool, token, block),
        token.price(block, skip_cache=skip_cache, sync=False),
    )
    return UsdValue(balance * Decimal(price))



import math
from collections import defaultdict
from decimal import Decimal
from functools import cached_property, lru_cache
from itertools import cycle, islice
from logging import DEBUG, getLogger
from typing import (
    AsyncIterator,
    DefaultDict,
    Dict,
    Iterable,
    List,
    Literal,
    Optional,
    Tuple,
    Union,
)

import a_sync
import eth_retry
from a_sync import igather
from a_sync.a_sync import HiddenMethodDescriptor
from brownie.network.event import _EventItem
from eth_typing import ChecksumAddress, HexAddress
from typing_extensions import Self

from y import convert
from y import ENVIRONMENT_VARIABLES as ENVS
from y._decorators import stuck_coro_debugger
from y.classes.common import ERC20, ContractBase
from y.constants import CHAINID, CONNECTED_TO_MAINNET, usdc, weth
from y.contracts import Contract, contract_creation_block_async
from y.datatypes import Address, AnyAddressType, Block, Pool, UsdPrice
from y.exceptions import (
    ContractNotVerified,
    TokenNotFound,
    call_reverted,
)
from y.interfaces.uniswap.quoterv3 import UNIV3_QUOTER_ABI
from y.networks import Network
from y.utils.events import ProcessedEvents

try:
    from eth_abi.packed import encode_packed
except ImportError:
    from eth_abi.packed import encode_abi_packed as encode_packed

# https://github.com/Uniswap/uniswap-v3-periphery/blob/main/deploys.md
UNISWAP_V3_FACTORY = "0x1F98431c8aD98523631AE4a59f267346ea31F984"
UNISWAP_V3_QUOTER = "0xb27308f9F90D607463bb33eA1BeBb41C27CE5AB6"

logger = getLogger(__name__)

Path = Iterable[Union[Address, int]]

# same addresses on all networks
addresses = {
    Network.Mainnet: {
        "factory": UNISWAP_V3_FACTORY,
        "quoter": UNISWAP_V3_QUOTER,
        "fee_tiers": (3000, 500, 10_000, 100),
    },
    Network.Arbitrum: {
        "factory": UNISWAP_V3_FACTORY,
        "quoter": UNISWAP_V3_QUOTER,
        "fee_tiers": (3000, 500, 10_000),
    },
    Network.Optimism: {
        "factory": UNISWAP_V3_FACTORY,
        "quoter": UNISWAP_V3_QUOTER,
        "fee_tiers": (3000, 500, 10_000, 100),
    },
    Network.Base: {
        "factory": "0x33128a8fC17869897dcE68Ed026d694621f6FDfD",
        "quoter": "0x3d4e44Eb1374240CE5F1B871ab261CD16335B76a",  # quoter v2
        "fee_tiers": (3000, 500, 10_000, 100),
    },
}

forked_deployments = {
    Network.Optimism: [
        {
            # Velodrome slipstream
            "factory": "0xCc0bDDB707055e04e497aB22a59c2aF4391cd12F",
            "quoter": "0x89D8218ed5fF1e46d8dcd33fb0bbeE3be1621466",
            "fee_tiers": (3000, 500, 10_000, 100),
        }
    ],
    Network.Base: [
        {
            # Aerodrome slipstream
            "factory": "0x5e7BB104d84c7CB9B682AaC2F3d509f5F406809A",
            "quoter": "0x254cF9E1E6e233aa1AC962CB9B05b2cfeAaE15b0",
            "fee_tiers": (3000, 500, 10_000, 100),
        },
    ],
}

_FEE_DENOMINATOR = Decimal(1_000_000)

_PATH_TYPE_STRINGS: Dict[int, Tuple[Literal["address", "uint24"], ...]] = {
    3: tuple(islice(cycle(("address", "uint24")), 3)),
    5: tuple(islice(cycle(("address", "uint24")), 5)),
}


class UniswapV3Pool(ContractBase):
    """Represents a Uniswap V3 Pool."""

    __contains_cache__: Dict[Address, Dict[ChecksumAddress, bool]] = {}

    __slots__ = "fee", "token0", "token1", "tick_spacing"

    def __init__(
        self,
        address: Address,
        token0: Address,
        token1: Address,
        tick_spacing: int,
        fee: int,
        deploy_block: int,
        asynchronous: bool = False,
    ) -> None:
        """
        Initialize a UniswapV3Pool instance.

        Args:
            address: The address of the pool.
            token0: The address of the first token in the pool.
            token1: The address of the second token in the pool.
            tick_spacing: The tick spacing of the pool.
            fee: The fee of the pool.
            deploy_block: The block number when the pool was deployed.
            asynchronous: Whether to use asynchronous operations.

        Examples:
            >>> pool = UniswapV3Pool(
            ...     address="0x1234567890abcdef1234567890abcdef12345678",
            ...     token0="0xToken0Address",
            ...     token1="0xToken1Address",
            ...     tick_spacing=60,
            ...     fee=3000,
            ...     deploy_block=1234567,
            ...     asynchronous=True
            ... )
        """
        super().__init__(address, asynchronous=asynchronous)
        self.token0 = ERC20(token0, asynchronous=asynchronous)
        self.token1 = ERC20(token1, asynchronous=asynchronous)
        self.tick_spacing = tick_spacing
        self.fee = fee
        self._deploy_block = deploy_block

    def __contains__(self, token: Address) -> bool:
        """
        Check if a token is part of the pool.

        Args:
            token: The address of the token to check.

        Returns:
            True if the token is part of the pool, False otherwise.

        Examples:
            >>> pool = UniswapV3Pool(...)
            >>> "0xToken0Address" in pool
            True
        """
        # force token to string in case it is Contract or EthAddress etc
        token = str(token)
        cache_for_token = self.__contains_cache__.get(token, {})
        cache_value = cache_for_token.get(self.address)
        if cache_value is None:
            if not cache_for_token:
                self.__contains_cache__[token] = {}
            cache_value = token in (self.token0.address, self.token1.address)
            self.__contains_cache__[token][self.address] = cache_value
        return cache_value

    def __getitem__(self, token: Address) -> ERC20:
        """
        Get the ERC20 token object for a given token address.

        Args:
            token: The address of the token.

        Returns:
            The ERC20 token object.

        Raises:
            TokenNotFound: If the token is not part of the pool.

        Examples:
            >>> pool = UniswapV3Pool(...)
            >>> token = pool["0xToken0Address"]
        """
        if token not in self:
            raise TokenNotFound(token, self)
        return ERC20(token, asynchronous=self.asynchronous)

    # DEBUG: lets try a semaphore here
    @a_sync.a_sync(ram_cache_maxsize=100_000, ram_cache_ttl=60 * 60, semaphore=10000)
    async def check_liquidity(
        self, token: AnyAddressType, block: Block
    ) -> Optional[int]:
        """
        Check the liquidity of a token in the pool at a specific block.

        Args:
            token: The address of the token.
            block: The block number to check liquidity at.

        Returns:
            The liquidity of the token in the pool, or None if not available.

        Examples:
            >>> pool = UniswapV3Pool(...)
            >>> liquidity = await pool.check_liquidity("0xToken0Address", 1234567)
        """
        if debug_logs_enabled := logger.isEnabledFor(DEBUG):
            logger._log(
                DEBUG,
                "checking %s liquidity for %s %s at %s",
                (
                    repr(self),
                    await ERC20(token, asynchronous=True).symbol,
                    token,
                    block,
                ),
            )

        if block < await self.deploy_block(sync=False):
            if debug_logs_enabled:
                logger._log(
                    DEBUG, "block %s prior to %s deploy block", (block, repr(self))
                )
            return 0

        try:
            liquidity = await self[token].balance_of(self.address, block, sync=False)
        except ContractNotVerified:
            if debug_logs_enabled:
                logger._log(
                    DEBUG,
                    "%s is not verified and we cannot fetch balance the usual way. returning 0.",
                    (token,),
                )
            return 0

        if debug_logs_enabled:
            await log_liquidity(self, token, block, liquidity)
        return liquidity

    @a_sync.a_sync(ram_cache_maxsize=100_000, ram_cache_ttl=60 * 60)
    async def _check_liquidity_token_out(
        self, token_in: AnyAddressType, block: Block
    ) -> Optional[int]:
        """
        Check the liquidity of the token out for a given token in.

        Args:
            token_in: The address of the token in.
            block: The block number to check liquidity at.

        Returns:
            The liquidity of the token out, or None if not available.

        Examples:
            >>> pool = UniswapV3Pool(...)
            >>> liquidity = await pool._check_liquidity_token_out("0xToken0Address", 1234567)
        """
        return await self.check_liquidity(
            self._get_token_out(token_in), block=block, sync=False
        )

    @lru_cache
    def _get_token_out(self, token_in: ERC20) -> ERC20:
        """
        Get the token out for a given token in.

        Args:
            token_in: The ERC20 token object for the token in.

        Returns:
            The ERC20 token object for the token out.

        Raises:
            TokenNotFound: If the token in is not part of the pool.

        Examples:
            >>> pool = UniswapV3Pool(...)
            >>> token_out = pool._get_token_out(token_in)
        """
        if token_in == self.token0:
            return self.token1
        elif token_in == self.token1:
            return self.token0
        raise TokenNotFound(token_in, self)


class UniswapV3(a_sync.ASyncGenericBase):
    """Represents the Uniswap V3 protocol."""

    def __init__(
        self,
        factory: HexAddress,
        quoter: HexAddress,
        fee_tiers: List[int],
        asynchronous: bool = True,
    ) -> None:
        """
        Initialize a UniswapV3 instance.

        Args:
            asynchronous: Whether to use asynchronous operations.

        Raises:
            UnsupportedNetwork: If Uniswap V3 is not supported on the current network.

        Examples:
            >>> uniswap_v3 = UniswapV3(asynchronous=True)
        """
        super().__init__()
        self.asynchronous = asynchronous
        self._factory = convert.to_address(factory)
        self._quoter = convert.to_address(quoter)
        self.fee_tiers = fee_tiers
        self.loading = False
        self._pools = {}

    def __repr__(self) -> str:
        return f"<{type(self).__name__} factory={self._factory} quoter={self._quoter}>"

    def __contains__(self, asset) -> bool:
        """
        Check if an asset is part of the Uniswap V3 protocol.

        Args:
            asset: The asset to check.

        Returns:
            True if the asset is part of the protocol, False otherwise.

        Examples:
            >>> uniswap_v3 = UniswapV3(...)
            >>> "0xAssetAddress" in uniswap_v3
            True
        """
        return CHAINID in addresses

    @cached_property
    def loaded(self) -> a_sync.Event:
        """
        Get the loaded event for the Uniswap V3 instance.

        Returns:
            The loaded event.

        Examples:
            >>> uniswap_v3 = UniswapV3(...)
            >>> loaded_event = uniswap_v3.loaded
        """
        return a_sync.Event(name=str(self))

    @a_sync.aka.property
    async def factory(self) -> Contract:
        """
        Get the factory contract for the Uniswap V3 protocol.

        Returns:
            The factory contract.

        Examples:
            >>> uniswap_v3 = UniswapV3(...)
            >>> factory_contract = await uniswap_v3.factory
        """
        return await Contract.coroutine(self._factory)

    __factory__: HiddenMethodDescriptor[Self, Contract]

    @a_sync.aka.cached_property
    async def quoter(self) -> Contract:
        """
        Get the quoter contract for the Uniswap V3 protocol.

        Returns:
            The quoter contract.

        Examples:
            >>> uniswap_v3 = UniswapV3(...)
            >>> quoter_contract = await uniswap_v3.quoter
        """
        try:
            return await Contract.coroutine(self._quoter)
        except ContractNotVerified:
            return Contract.from_abi("Quoter", self._quoter, UNIV3_QUOTER_ABI)

    __quoter__: HiddenMethodDescriptor[Self, Contract]

    @a_sync.aka.cached_property
    @stuck_coro_debugger
    async def pools(self) -> List[UniswapV3Pool]:
        """
        Get the list of Uniswap V3 pools.

        Returns:
            A list of Uniswap V3 pools.

        Examples:
            >>> uniswap_v3 = UniswapV3(...)
            >>> pools = await uniswap_v3.pools
        """
        factory = await self.__factory__
        if (
            CHAINID == Network.Base
            and factory.address == "0x5e7BB104d84c7CB9B682AaC2F3d509f5F406809A"
        ):
            return SlipstreamPools(factory, asynchronous=self.asynchronous)
        elif (
            CHAINID == Network.Optimism
            and factory.address == "0xCc0bDDB707055e04e497aB22a59c2aF4391cd12F"
        ):
            return SlipstreamPools(factory, asynchronous=self.asynchronous)
        return UniV3Pools(factory, asynchronous=self.asynchronous)

    __pools__: HiddenMethodDescriptor[Self, "UniV3Pools"]

    async def pools_for_token(
        self, token: Address, block: Block
    ) -> AsyncIterator[UniswapV3Pool]:
        """
        Get the pools for a specific token.

        Args:
            token: The address of the token.
            block: The block number to get pools at.

        Yields:
            Uniswap V3 pools for the token.

        Examples:
            >>> uniswap_v3 = UniswapV3(...)
            >>> async for pool in uniswap_v3.pools_for_token("0xTokenAddress", 1234567):
            ...     print(pool)
        """
        pools = await self.__pools__
        async for pool in pools.objects(to_block=block):
            if token in pool:
                yield pool

    @stuck_coro_debugger
    @a_sync.a_sync(cache_type="memory", ram_cache_ttl=ENVS.CACHE_TTL)
    async def get_price(
        self,
        token: Address,
        block: Optional[Block] = None,
        ignore_pools: Tuple[Pool, ...] = (),  # unused
        skip_cache: bool = ENVS.SKIP_CACHE,  # unused
    ) -> Optional[UsdPrice]:
        """
        Get the price of a token in USD.

        Args:
            token: The address of the token.
            block: The block number to get the price at.
            ignore_pools: Pools to ignore (unused).
            skip_cache: Whether to skip cache (unused).

        Returns:
            The price of the token in USD, or None if not available.

        Examples:
            >>> uniswap_v3 = UniswapV3(...)
            >>> price = await uniswap_v3.get_price("0xTokenAddress", 1234567)
        """
        quoter = await self.__quoter__
        if block and block < await contract_creation_block_async(quoter, True):
            return None

        paths: List[Path] = [(token, fee, usdc.address) for fee in self.fee_tiers]
        if token != weth:
            paths += [
                (token, fee, weth.address, self.fee_tiers[0], usdc.address)
                for fee in self.fee_tiers
            ]

        if debug_logs_enabled := logger.isEnabledFor(DEBUG):
            logger._log(DEBUG, "paths: %s", (paths,))

        amount_in = await ERC20._get_scale_for(token)
        results = await igather(
            self._quote_exact_input(path, amount_in, block) for path in paths
        )

        if debug_logs_enabled:
            logger._log(DEBUG, "results: %s", (results,))

        outputs = list(filter(None, results))
        if debug_logs_enabled:
            logger._log(DEBUG, "outputs: %s", (outputs,))
        return UsdPrice(max(outputs)) if outputs else None

    @stuck_coro_debugger
    @a_sync.a_sync(ram_cache_maxsize=100_000, ram_cache_ttl=60 * 60)
    async def check_liquidity(
        self, token: Address, block: Block, ignore_pools: Tuple[Pool, ...] = ()
    ) -> int:
        """
        Check the liquidity of a token in the Uniswap V3 protocol.

        Args:
            token: The address of the token.
            block: The block number to check liquidity at.
            ignore_pools: Pools to ignore.

        Returns:
            The liquidity of the token.

        Examples:
            >>> uniswap_v3 = UniswapV3(...)
            >>> liquidity = await uniswap_v3.check_liquidity("0xTokenAddress", 1234567)
        """
        if debug_logs_enabled := logger.isEnabledFor(DEBUG):
            logger._log(
                DEBUG,
                "checking %s liquidity for %s %s at %s",
                (self, await ERC20(token, asynchronous=True).symbol, token, block),
            )

        if (
            CONNECTED_TO_MAINNET
            and token == "0x6DEA81C8171D0bA574754EF6F8b412F2Ed88c54D"
        ):
            # LQTY, TODO refactor this somehow
            return 0

        quoter = await self.__quoter__
        if block and block < await contract_creation_block_async(quoter):
            if debug_logs_enabled:
                logger._log(
                    DEBUG, "block %s is before %s deploy block", (block, quoter)
                )
            return 0

        if token == weth.address:
            # NOTE: we need to filter these or else we will be fetching every pool
            #       for now, we only focus on weth/usdc pools
            filter_fn = (
                lambda pool: pool._get_token_out(token) == usdc.address
                and pool not in ignore_pools
            )
        else:
            filter_fn = lambda pool: pool not in ignore_pools

        token_out_tasks = UniswapV3Pool._check_liquidity_token_out.map(
            token_in=token, block=block
        )
        pools_for_token: a_sync.ASyncIterator[UniswapV3Pool] = self.pools_for_token(
            token, block
        )

        if debug_logs_enabled:
            async for pool in pools_for_token.filter(filter_fn):
                # the mapping will start the tasks internally
                logger._log(DEBUG, "starting token_out_task for %s", (pool,))
                token_out_tasks[pool]
        else:
            async for pool in pools_for_token.filter(filter_fn):
                token_out_tasks[pool]

        if not token_out_tasks:
            return 0

        if debug_logs_enabled:
            logger._log(
                DEBUG,
                "%s token_out_tasks for %s at %s: %s",
                (self, token, block, token_out_tasks),
            )

        # Since uni v3 liquidity can be provided asymmetrically, the most liquid pool in terms of `token` might not actually be the most liquid pool in terms of `token_out`
        # We need some spaghetticode here to account for these erroneous liquidity values
        # TODO: Refactor this
        token_out_liquidity: DefaultDict[ERC20, List[int]] = defaultdict(list)
        if debug_logs_enabled:
            async for pool, liquidity in token_out_tasks.map(pop=False):
                await log_liquidity(pool, token, block, liquidity)
                token_out_liquidity[pool._get_token_out(token)].append(liquidity)
            logger._log(
                DEBUG, "%s token_out_liquidity: %s", (token, token_out_liquidity)
            )
        else:
            async for pool, liquidity in token_out_tasks.map(pop=False):
                token_out_liquidity[pool._get_token_out(token)].append(liquidity)

        token_out_min_liquidity = {
            token_out: min(liquidities)
            for token_out, liquidities in token_out_liquidity.items()
        }

        token_in_tasks = UniswapV3Pool.check_liquidity.map(token=token, block=block)
        async for pool, liquidity in token_out_tasks.map(pop=True):
            token_out = pool._get_token_out(token)
            if (
                len(token_out_liquidity[token_out]) > 1
                and liquidity == token_out_min_liquidity[token_out]
            ):
                if debug_logs_enabled:
                    logger._log(DEBUG, "ignoring liquidity for %s", (pool,))
            elif token_out == weth and liquidity < 10**19:  # 10 ETH
                # NOTE: this is totally arbitrary, works for all known cases but eventually will probably cause issues
                if debug_logs_enabled:
                    logger._log(DEBUG, "insufficient liquidity for %s", (pool,))
            else:
                token_in_tasks[pool]

        liquidity = (
            await token_in_tasks.max(pop=True, sync=False) if token_in_tasks else 0
        )
        if debug_logs_enabled:
            await log_liquidity(self, token, block, liquidity)
        return liquidity

    @stuck_coro_debugger
    @eth_retry.auto_retry
    async def _quote_exact_input(
        self, path: Path, amount_in: int, block: int
    ) -> Optional[Decimal]:
        """
        Quote the exact input for a given path and amount.

        Args:
            path: The path for the swap.
            amount_in: The input amount.
            block: The block number to quote at.

        Returns:
            The quoted output amount.

        Examples:
            >>> uniswap_v3 = UniswapV3(...)
            >>> output_amount = await uniswap_v3._quote_exact_input(path, 1000, 1234567)
        """
        quoter = await self.__quoter__
        try:
            amount = await quoter.quoteExactInput.coroutine(
                _encode_path(path), amount_in, block_identifier=block
            )
            return (
                # Quoter v2 uses this weird return struct, we must unpack it to get amount out.
                (amount if isinstance(amount, int) else amount[0])
                / _undo_fees(path)
                / _FEE_DENOMINATOR
            )
        except Exception as e:
            if not call_reverted(e):
                raise


def _encode_path(path: Path) -> bytes:
    """
    Encode a path for Uniswap V3.

    Args:
        path: The path to encode.

    Returns:
        The encoded path.

    Examples:
        >>> path = ["0xToken0Address", 3000, "0xToken1Address"]
        >>> encoded_path = _encode_path(path)
    """
    return encode_packed(_PATH_TYPE_STRINGS[len(path)], path)


def _undo_fees(path: Path) -> Decimal:
    """
    Undo the fees for a given path.

    Args:
        path: The path to undo fees for.

    Returns:
        The fee multiplier.

    Examples:
        >>> path = ["0xToken0Address", 3000, "0xToken1Address"]
        >>> fee_multiplier = _undo_fees(path)
    """
    fees = (1 - fee / _FEE_DENOMINATOR for fee in islice(path, 1, None, 2))
    return math.prod(fees)


class UniV3Pools(ProcessedEvents[UniswapV3Pool]):
    """Represents a collection of Uniswap V3 Pools."""

    __slots__ = ("asynchronous",)

    def __init__(self, factory: Contract, asynchronous: bool = False):
        """
        Initialize a UniV3Pools instance.

        Args:
            factory: The factory contract for Uniswap V3.
            asynchronous: Whether to use asynchronous operations.

        Examples:
            >>> factory_contract = Contract(...)
            >>> pools = UniV3Pools(factory_contract, asynchronous=True)
        """
        self.asynchronous = asynchronous
        super().__init__(
            addresses=[factory.address], topics=[factory.topics["PoolCreated"]]
        )

    def _process_event(self, event: _EventItem) -> UniswapV3Pool:
        """
        Process a PoolCreated event and return a UniswapV3Pool instance.

        Args:
            event: The PoolCreated event.

        Returns:
            A UniswapV3Pool instance.

        Examples:
            >>> pools = UniV3Pools(...)
            >>> pool = pools._process_event(event)
        """
        token0, token1, fee, tick_spacing, pool = event.values()
        return UniswapV3Pool(
            pool,
            token0,
            token1,
            fee,
            tick_spacing,
            event.block_number,
            asynchronous=self.asynchronous,
        )

    def _get_block_for_obj(self, obj: UniswapV3Pool) -> int:
        """
        Get the block number for a UniswapV3Pool object.

        Args:
            obj: The UniswapV3Pool object.

        Returns:
            The block number.

        Examples:
            >>> pools = UniV3Pools(...)
            >>> block_number = pools._get_block_for_obj(pool)
        """
        return obj._deploy_block


class SlipstreamPools(UniV3Pools):
    def _process_event(self, event: _EventItem) -> UniswapV3Pool:
        token0, token1, tick_spacing, pool = event.values()
        return UniswapV3Pool(
            pool,
            token0,
            token1,
            # NOTE: fee arg is not actually used in the current implementation, so we can use 0 here
            0,  # TODO: implement fee maths properly
            tick_spacing,
            event.block_number,
            asynchronous=self.asynchronous,
        )


if CHAINID in addresses:
    uniswap_v3 = UniswapV3(
        addresses[CHAINID]["factory"],
        addresses[CHAINID]["quoter"],
        addresses[CHAINID]["fee_tiers"],
        asynchronous=True,
    )
else:
    uniswap_v3 = None

forks = [
    UniswapV3(fork["factory"], fork["quoter"], fork["fee_tiers"], asynchronous=True)
    for fork in forked_deployments.get(CHAINID, [])
]


async def log_liquidity(market, token, block, liquidity) -> None:
    __logger_log(
        DEBUG,
        "%s liquidity for %s %s at %s: %s",
        (
            repr(market),
            await ERC20(token, asynchronous=True).symbol,
            token,
            block,
            liquidity,
        ),
    )


__logger_log = logger._log



"""
This module provides utility functions for the ypricemagic library.

The utilities include caching, attribute checking, and method gathering functionalities
that are used throughout the library to enhance performance and ensure code reliability.

Imported Functions:
    - :func:`y.utils.cache.a_sync_ttl_cache`: Provides a caching mechanism with a time-to-live (TTL) feature.
    - :func:`y.utils.checks.hasall`: Checks if an object has all specified attributes.
    - :func:`y.utils.gather.gather_methods`: Gathers methods from a class or module.

Examples:
    Using the caching utility:
    >>> from y.utils.cache import a_sync_ttl_cache
    >>> @a_sync_ttl_cache
    ... def expensive_function(x):
    ...     return x * x
    >>> result = expensive_function(4)

    Checking for attributes:
    >>> from y.utils.checks import hasall
    >>> class MyClass:
    ...     attr1 = 1
    ...     attr2 = 2
    >>> obj = MyClass()
    >>> hasall(obj, ['attr1', 'attr2'])
    True

    Gathering methods:
    >>> from y.utils.gather import gather_methods
    >>> methods = gather_methods(MyClass)
    >>> print(methods)

See Also:
    - :mod:`y.utils.cache`: For caching utilities.
    - :mod:`y.utils.checks`: For attribute checking utilities.
    - :mod:`y.utils.gather`: For method gathering utilities.
"""

from y.utils.cache import a_sync_ttl_cache
from y.utils.checks import hasall
from y.utils.gather import gather_methods



import logging
from typing import Optional, Tuple

import a_sync

from y import ENVIRONMENT_VARIABLES as ENVS
from y import convert
from y.classes.common import ERC20, WeiBalance
from y.contracts import has_methods
from y.datatypes import AnyAddressType, Block, UsdPrice, UsdValue
from y.exceptions import call_reverted
from y.utils import gather_methods
from y.utils.cache import optional_async_diskcache

_RESERVES_METHODS = (
    "token0()(address)",
    "token1()(address)",
    "usersAmounts()((uint,uint))",
)

logger = logging.getLogger(__name__)


@a_sync.a_sync(default="sync", cache_type="memory", ram_cache_ttl=5 * 60)
@optional_async_diskcache
async def is_popsicle_lp(token_address: AnyAddressType) -> bool:
    """
    Determines if the given token address is a Popsicle Finance LP token.

    Args:
        token_address: The address of the token to check.

    Returns:
        True if the token is a Popsicle LP token, False otherwise.

    Example:
        >>> is_popsicle = is_popsicle_lp("0xd2C5A739ebfE3E00CFa88A51749d367d7c496CCf")
        >>> print(is_popsicle)
        True
    """
    # NOTE: contract to check for reference (mainnet): 0xd2C5A739ebfE3E00CFa88A51749d367d7c496CCf
    return await has_methods(
        token_address,
        ("token0()(address)", "token1()(address)", "usersAmounts()((uint,uint))"),
        sync=False,
    )


@a_sync.a_sync(default="sync")
async def get_price(
    token: AnyAddressType,
    block: Optional[Block] = None,
    *,
    skip_cache: bool = ENVS.SKIP_CACHE
) -> Optional[UsdPrice]:
    """
    Calculates the price of a Popsicle Finance LP token.

    Args:
        token: The address of the Popsicle LP token.
        block (optional): The block number to query. Defaults to the latest block.
        skip_cache (optional): Whether to skip the cache when fetching prices. Defaults to :obj:`ENVS.SKIP_CACHE`.

    Returns:
        The price of the LP token in USD, or None if the price cannot be determined.

    Example:
        >>> price = get_price("0xd2C5A739ebfE3E00CFa88A51749d367d7c496CCf", block=14_000_000)
        >>> print(f"{price:.6f}")
        1.234567  # The price of the Popsicle LP token in USD

    See Also:
        - :func:`get_tvl`
        - :func:`get_balances`
    """
    address = await convert.to_address_async(token)
    total_val = await get_tvl(address, block, skip_cache=skip_cache, sync=False)
    if total_val is None:
        return None
    total_supply = await ERC20(address, asynchronous=True).total_supply_readable(
        block, sync=False
    )
    return UsdPrice(total_val / total_supply)


@a_sync.a_sync(default="sync")
async def get_tvl(
    token: AnyAddressType,
    block: Optional[Block] = None,
    skip_cache: bool = ENVS.SKIP_CACHE,
) -> Optional[UsdValue]:
    """
    Calculates the total value locked (TVL) in a Popsicle Finance LP token.

    Args:
        token: The address of the Popsicle LP token.
        block (optional): The block number to query. Defaults to the latest block.
        skip_cache (optional): Whether to skip the cache when fetching TVL. Defaults to :obj:`ENVS.SKIP_CACHE`.

    Returns:
        The total value locked in the LP token in USD, or None if the TVL cannot be determined.

    Example:
        >>> tvl = get_tvl("0xd2C5A739ebfE3E00CFa88A51749d367d7c496CCf", block=14_000_000)
        >>> print(f"{tvl:.2f}")
        1234567.89  # The TVL of the Popsicle LP token in USD

    See Also:
        - :func:`get_price`
        - :func:`get_balances`
    """
    balances: Tuple[WeiBalance, WeiBalance]
    balances = await get_balances(
        token, block, skip_cache=skip_cache, _async_balance_objects=True, sync=False
    )
    return UsdValue(await WeiBalance.value_usd.sum(balances)) if balances else None


@a_sync.a_sync(default="sync")
async def get_balances(
    token: AnyAddressType,
    block: Optional[Block] = None,
    skip_cache: bool = ENVS.SKIP_CACHE,
    _async_balance_objects: bool = False,
) -> Optional[Tuple[WeiBalance, WeiBalance]]:
    """
    Retrieves the balances of the underlying tokens in a Popsicle Finance LP token.

    Args:
        token: The address of the Popsicle LP token.
        block (optional): The block number to query. Defaults to the latest block.
        skip_cache (optional): Whether to skip the cache when fetching balances. Defaults to :obj:`ENVS.SKIP_CACHE`.
        _async_balance_objects (optional): Whether to use asynchronous balance objects. Defaults to False.

    Returns:
        A tuple containing the balances of the two underlying tokens, or None if the balances cannot be determined.

    Example:
        >>> balances = get_balances("0xd2C5A739ebfE3E00CFa88A51749d367d7c496CCf", block=14_000_000)
        >>> if balances:
        ...     balance0, balance1 = balances
        ...     print(f"Token0 Balance: {balance0.readable}, Token1 Balance: {balance1.readable}")
        ... else:
        ...     print("Balances could not be determined.")
        Token0 Balance: 1000.0, Token1 Balance: 2000.0

    Note:
        If the contract probed has the correct method name but the method returns the wrong type for Popsicle,
        an exception "not enough values to unpack (expected 3, got 2)" will be caught, and the function will return None.

    See Also:
        - :func:`get_price`
        - :func:`get_tvl`
    """
    try:
        token0, token1, (balance0, balance1) = await gather_methods(
            await convert.to_address_async(token), _RESERVES_METHODS, block=block
        )
    except Exception as e:
        if call_reverted(e):
            return None
        elif str(e) == "not enough values to unpack (expected 3, got 2)":
            # TODO determine if this is regular behavior when no tvl in pool or if this is bug to fix
            return None
        raise
    balance0 = WeiBalance(
        balance0,
        token0,
        block=block,
        skip_cache=skip_cache,
        asynchronous=_async_balance_objects,
    )
    balance1 = WeiBalance(
        balance1,
        token1,
        block=block,
        skip_cache=skip_cache,
        asynchronous=_async_balance_objects,
    )
    return balance0, balance1



from asyncio import create_task, iscoroutinefunction, sleep
from functools import wraps
from inspect import isasyncgenfunction
from logging import DEBUG, getLogger
from time import time
from typing import (
    AsyncIterator,
    Awaitable,
    Callable,
    NoReturn,
    TypeVar,
    Union,
    overload,
)

from a_sync import ASyncGenericBase
from a_sync.a_sync.method import ASyncBoundMethod
from a_sync.iter import ASyncGeneratorFunction
from eth_utils.toolz import valmap
from typing_extensions import Concatenate, ParamSpec

P = ParamSpec("P")
T = TypeVar("T")


def continue_on_revert(func: Callable[P, T]) -> Callable[P, T]:
    """
    Decorates a call-making function. If the call reverts, it attempts to continue
    by calling the standalone function :func:`continue_if_call_reverted` from the
    `y.exceptions` module. Raises any other exceptions.

    This decorator is useful for functions interacting with smart contracts where
    a call might revert due to various reasons. It ensures that the function does not
    fail completely on a revert, but instead handles it gracefully.

    Note:
        The decorated function will return `None` if the call reverts and
        :func:`continue_if_call_reverted` does not raise an exception.

    Args:
        func: The function to be decorated.

    Returns:
        The wrapped function with revert handling.

    Examples:
        >>> @continue_on_revert
        ... def fetch_data():
        ...     # some logic that might revert
        ...     pass

        >>> @continue_on_revert
        ... async def fetch_data_async():
        ...     # some async logic that might revert
        ...     pass

    See Also:
        - :func:`continue_if_call_reverted`
    """
    from y.exceptions import continue_if_call_reverted

    if iscoroutinefunction(func):

        @wraps(func)
        async def continue_on_revert_wrap(
            *args: P.args, **kwargs: P.kwargs
        ) -> Union[T, None]:
            try:
                return await func(*args, **kwargs)
            except Exception as e:
                continue_if_call_reverted(e)

    elif callable(func):

        @wraps(func)
        def continue_on_revert_wrap(
            *args: P.args, **kwargs: P.kwargs
        ) -> Union[T, None]:
            try:
                return func(*args, **kwargs)
            except Exception as e:
                continue_if_call_reverted(e)

    else:
        raise NotImplementedError(f"Unable to decorate {func}")
    return continue_on_revert_wrap


B = TypeVar("B", bound=ASyncGenericBase)


stuck_coro_logger = getLogger("y.stuck?")
__stuck_coro_logger_is_enabled_for = stuck_coro_logger.isEnabledFor
__stuck_coro_logger_log = stuck_coro_logger._log


@overload
def stuck_coro_debugger(
    fn: Callable[Concatenate[B, P], AsyncIterator[T]],
) -> ASyncGeneratorFunction[P, T]: ...


@overload
def stuck_coro_debugger(
    fn: Callable[Concatenate[B, P], Awaitable[T]],
) -> ASyncBoundMethod[B, P, T]: ...


@overload
def stuck_coro_debugger(
    fn: Callable[Concatenate[B, P], T],
) -> ASyncBoundMethod[B, P, T]: ...


@overload
def stuck_coro_debugger(
    fn: Callable[P, AsyncIterator[T]],
) -> Callable[P, AsyncIterator[T]]: ...


@overload
def stuck_coro_debugger(fn: Callable[P, Awaitable[T]]) -> Callable[P, Awaitable[T]]: ...


def stuck_coro_debugger(fn):
    if isasyncgenfunction(fn):

        @wraps(fn)
        async def stuck_async_gen_wrap(
            *args: P.args, **kwargs: P.kwargs
        ) -> AsyncIterator[T]:
            aiterator = fn(*args, **kwargs)

            if not __stuck_coro_logger_is_enabled_for(DEBUG):
                async for thing in aiterator:
                    yield thing
                return

            task = create_task(
                coro=_stuck_debug_task(fn, args, kwargs),
                name="_stuck_debug_task",
            )
            try:
                async for thing in aiterator:
                    yield thing
            finally:
                task.cancel()

        return stuck_async_gen_wrap
    else:

        @wraps(fn)
        async def stuck_coro_wrap(*args: P.args, **kwargs: P.kwargs) -> T:
            if not __stuck_coro_logger_is_enabled_for(DEBUG):
                return await fn(*args, **kwargs)

            task = create_task(
                coro=_stuck_debug_task(fn, args, kwargs),
                name="_stuck_debug_task",
            )
            try:
                retval = await fn(*args, **kwargs)
            finally:
                task.cancel()
            return retval

        return stuck_coro_wrap


async def _stuck_debug_task(
    fn: Callable[P, T], *args: P.args, **kwargs: P.kwargs
) -> NoReturn:
    # sleep early so fast-running coros can exit early
    await sleep(300)

    start = time() - 300
    module = fn.__module__
    name = fn.__name__
    formatted_args = tuple(map(str, args))
    formatted_kwargs = valmap(str, kwargs)
    while True:
        __stuck_coro_logger_log(
            DEBUG,
            "%s.%s still executing after %sm with args %s kwargs %s",
            (
                module,
                name,
                round((time() - start) / 60, 2),
                formatted_args,
                formatted_kwargs,
            ),
        )
        await sleep(300)



import pytest
from y.constants import STABLECOINS
from y.prices import magic

from tests.fixtures import blocks_for_contract

STABLECOINS = STABLECOINS.keys()


# TODO Implement this after implementing non-stable stables to validate tokens' spots in `STABLECOINS`.
@pytest.mark.parametrize("token", STABLECOINS)
def test_stablecoins(token):
    """
    Placeholder test for stablecoin prices.

    This test is intended to validate that the price of each stablecoin in the `STABLECOINS` list is $1.
    However, it is currently a placeholder and not fully implemented. The test will iterate over a set
    of blocks for each stablecoin contract and assert that the price is $1.

    Args:
        token: The stablecoin token address to be tested.

    See Also:
        - :func:`y.prices.magic.get_price` for retrieving token prices.
        - :func:`tests.fixtures.blocks_for_contract` for generating block numbers for testing.
    """
    for block in blocks_for_contract(token, 20):
        # NOTE Placeholder.
        assert (
            magic.get_price(token, block, skip_cache=True) == 1
        ), "Stablecoin price not $1"



import logging
import threading
from typing import Dict, Optional

from a_sync import ProcessingQueue, a_sync
from cachetools import TTLCache, cached
from pony.orm import select

from y._db.decorators import db_session_retry_locked, log_result_count
from y._db.entities import Contract
from y._db.utils._ep import _get_get_token
from y._db.utils.utils import ensure_block, make_executor
from y.constants import CHAINID
from y.datatypes import Address, Block


logger = logging.getLogger(__name__)
_logger_debug = logger.debug

_get_get_token()

_deploy_block_read_executor = make_executor(
    2, 4, "ypricemagic db executor [deploy block]"
)
_deploy_block_write_executor = make_executor(
    1, 4, "ypricemagic db executor [deploy block]"
)


@a_sync(
    default="async",
    executor=_deploy_block_read_executor,
)
@db_session_retry_locked
def get_deploy_block(address: str) -> Optional[int]:
    """Retrieve the deployment block number for a given contract address.

    This function first checks a cache for the deployment block number. If not found,
    it fetches the information from the blockchain.

    Args:
        address: The contract address as a string.

    Examples:
        >>> deploy_block = get_deploy_block("0x1234567890abcdef1234567890abcdef12345678")
        >>> print(deploy_block)

    See Also:
        - :func:`_set_deploy_block`
    """
    if deploy_block := known_deploy_blocks().pop(address, None):
        _logger_debug("%s deploy block from cache: %s", address, deploy_block)
        return deploy_block
    get_token = _get_get_token()
    token = get_token(address, sync=True)
    if token is None:
        raise ValueError(f"get_token('{address}', sync=True) returned 'None'")
    if deploy_block := token.deploy_block:
        _logger_debug("%s deploy block from cache: %s", address, deploy_block.number)
        return deploy_block.number
    _logger_debug("%s deploy block not cached, fetching from chain", address)


@a_sync(default="async", executor=_deploy_block_write_executor)
@db_session_retry_locked
def _set_deploy_block(address: str, deploy_block: int) -> None:
    """Set the deployment block number for a contract address in the database.

    Args:
        address: The contract address as a string.
        deploy_block: The block number where the contract was deployed.

    See Also:
        - :func:`set_deploy_block`
    """
    from y._db.utils._ep import _get_get_token

    ensure_block(deploy_block, sync=True)
    get_token = _get_get_token()
    get_token(address, sync=True).deploy_block = (CHAINID, deploy_block)
    _logger_debug("deploy block cached for %s: %s", address, deploy_block)


def set_deploy_block(address: str, deploy_block: int) -> None:
    """Set the deployment block number for a contract address in the database.

    Args:
        address: The contract address as a string.
        deploy_block: The block number where the contract was deployed.

    See Also:
        - :func:`_set_deploy_block`
    """


set_deploy_block = ProcessingQueue(_set_deploy_block, num_workers=10, return_data=False)

# startup caches


@cached(TTLCache(maxsize=1, ttl=60 * 60), lock=threading.Lock())
@log_result_count("deploy blocks")
def known_deploy_blocks() -> Dict[Address, Block]:
    """Cache and return all known contract deploy blocks for this chain.

    This function minimizes database reads by caching the result for one hour.

    Examples:
        >>> deploy_blocks = known_deploy_blocks()
        >>> print(deploy_blocks)

    See Also:
        - :func:`get_deploy_block`
    """
    return dict(
        select(
            (c.address, c.deploy_block.number)
            for c in Contract
            if c.chain.id == CHAINID and c.deploy_block.number
        )
    )



import a_sync
import pytest
from brownie import chain

from tests.fixtures import blocks_for_contract
from y.constants import WRAPPED_GAS_COIN
from y.networks import Network
from y.prices import popsicle

POPSICLES = {
    Network.Mainnet: [
        "0xa7053782dC3523D2C82B439Acf3f9344Fb47b97f",
        "0xd2C5A739ebfE3E00CFa88A51749d367d7c496CCf",
        "0x949FDF28F437258E7564a35596b1A99b24F81e4e",
        "0xa0273C10b8A4BF0bDC57cb0bC974E3A9d89527b8",
        "0x495410B129A27bC771ce8fb316d804a5686B8Ea7",
        "0x5C08A6762CAF9ec8a42F249eBC23aAE66097218D",
        "0xaE7b92C8B14E7bdB523408aE0A6fFbf3f589adD9",
        "0x9683D433621A83aA7dd290106e1da85251317F55",
        "0x212Aa024E25A9C9bAF5b5397B558B7ccea81740B",
        "0xBE5d1d15617879B22C7b6a8e1e16aDD6d0bE3c61",
        "0xFF338D347E59d6B61E5C69382915D863bb22Ef2f",
        "0xa1BE64Bb138f2B6BCC2fBeCb14c3901b63943d0E",
        "0x8d8B490fCe6Ca1A31752E7cFAFa954Bf30eB7EE2",
        "0x989442D5cCB27E7931095B0f3165c75a6def9bc3",
        "0xbA38029806AbE4B45D5273098137DDb52dA8e62F",
        "0xd2EF15af2649CC46e3E23B96563a3d44ef5E5A06",
        "0xF4f542E4b5E2345A1f2D0fEab9492357Ebc5c8f4",
        "0x36e9B6e7FADC7b8Ee289c8A24Ad96573cda3D7D9",
    ],
    Network.Polygon: [
        "0x5C08A6762CAF9ec8a42F249eBC23aAE66097218D",
        "0xaE7b92C8B14E7bdB523408aE0A6fFbf3f589adD9",
        "0xa0273C10b8A4BF0bDC57cb0bC974E3A9d89527b8",
        "0x949FDF28F437258E7564a35596b1A99b24F81e4e",
        "0x9683D433621A83aA7dd290106e1da85251317F55",
        "0xd2C5A739ebfE3E00CFa88A51749d367d7c496CCf",
        "0xa7053782dC3523D2C82B439Acf3f9344Fb47b97f",
    ],
}.get(chain.id, [])


def test_non_popsicle():
    """Test that a non-Popsicle LP token is not incorrectly identified as a Popsicle LP.

    This test checks that the `WRAPPED_GAS_COIN` is not recognized as a Popsicle LP token
    and that attempting to get its price raises a `TypeError`.

    See Also:
        - :func:`popsicle.is_popsicle_lp`
        - :func:`popsicle.get_price`
    """
    assert not popsicle.is_popsicle_lp(
        WRAPPED_GAS_COIN
    ), "Token incorrectly recognized as Popsicle LP."
    with pytest.raises(TypeError):
        popsicle.get_price(WRAPPED_GAS_COIN)


@pytest.mark.asyncio_cooperative
@pytest.mark.parametrize("token", POPSICLES)
async def test_popsicle_get_price(token):
    """Test fetching the price of Popsicle LP tokens.

    This test verifies that each token in the `POPSICLES` list is correctly identified
    as a Popsicle LP and that its price can be fetched for various blocks.

    Args:
        token: The Popsicle LP token address to test.

    See Also:
        - :func:`popsicle.is_popsicle_lp`
        - :func:`popsicle.get_price`
        - :func:`blocks_for_contract`
    """
    assert await popsicle.is_popsicle_lp(
        token, sync=False
    ), "Popsicle LP not recognized."
    blocks = blocks_for_contract(token, 25)
    async for block, price in a_sync.map(popsicle.get_price, blocks, token=token):
        assert price, f"Failed to fetch price for {token} at block {block}."



from asyncio import as_completed, ensure_future, iscoroutine
from logging import DEBUG, getLogger
from typing import Awaitable, Callable, Tuple

import a_sync

from y import convert
from y.classes.common import ERC20
from y.constants import STABLECOINS
from y.datatypes import Address, AnyAddressType
from y.prices import convex, one_to_one, pendle, popsicle, rkp3r, solidex, yearn
from y.prices.band import band
from y.prices.chainlink import chainlink
from y.prices.dex import mooniswap
from y.prices.dex.balancer import balancer_multiplexer
from y.prices.dex.genericamm import is_generic_amm
from y.prices.dex.uniswap import uniswap_multiplexer
from y.prices.eth_derivs import creth, wsteth
from y.prices.gearbox import gearbox
from y.prices.lending import ib
from y.prices.lending.aave import aave
from y.prices.lending.compound import compound
from y.prices.stable_swap import belt, ellipsis, froyo, mstablefeederpool, saddle
from y.prices.stable_swap.curve import curve
from y.prices.synthetix import synthetix
from y.prices.tokenized_fund import basketdao, gelato, piedao, reserve, tokensets
from y.utils.logging import get_price_logger

logger = getLogger(__name__)


@a_sync.a_sync(default="sync", cache_type="memory")
async def check_bucket(token: AnyAddressType) -> str:
    """
    Determine the category or 'bucket' of a given token.

    This function attempts to classify a token into a specific category based on its characteristics.
    It first checks the database for a cached bucket using :func:`y._db.utils.token.get_bucket`.
    If not found, it performs a series of checks that may involve string comparisons, calls to the blockchain,
    and contract initializations. The determined bucket is then stored using :func:`db.set_bucket`.

    Args:
        token: The token address to classify.

    Examples:
        >>> bucket = check_bucket("0x6B175474E89094C44Da98b954EedeAC495271d0F")
        >>> print(bucket)
        'stable usd'

    See Also:
        - :func:`y.convert.to_address_async`
        - :func:`y.utils.logging.get_price_logger`
        - :func:`_check_bucket_helper`
    """
    token_address = await convert.to_address_async(token)
    logger = get_price_logger(token_address, block=None, extra="buckets")

    import y._db.utils.token as db

    bucket = await db.get_bucket(token_address)
    if bucket:
        logger.debug("returning bucket %s from ydb", bucket)
        return bucket

    debug_logs_enabled = logger.isEnabledFor(DEBUG)

    # these require neither calls to the chain nor contract initialization, just string comparisons (pretty sure)
    for bucket, check in string_matchers.items():
        if check(token):
            if debug_logs_enabled:
                await __log_bucket(token_address, bucket)
            db.set_bucket(token, bucket)
            return bucket
        elif debug_logs_enabled:
            await __log_not_bucket(token_address, bucket)

    # check these first, these just require calls
    futs = [
        ensure_future(_check_bucket_helper(bucket, check, token_address))
        for bucket, check in calls_only.items()
    ]
    for fut in as_completed(futs):
        try:
            bucket, is_member = await fut
        except TypeError:
            raise
        except Exception as e:
            logger.warning(
                "%s when checking %s. This will probably not impact your run.", e, fut
            )
            logger.warning(e, exc_info=True)
            continue

        if is_member:
            if debug_logs_enabled:
                await __log_bucket(token_address, bucket)
            for fut in futs:
                fut.cancel()
            db.set_bucket(token, bucket)
            return bucket
        else:
            if debug_logs_enabled:
                await __log_not_bucket(token_address, bucket)
            bucket = None

    # TODO: Refactor the below like the above
    # these require both calls and contract initializations
    if await solidex.is_solidex_deposit(token_address, sync=False):
        bucket = "solidex"
    if await uniswap_multiplexer.is_uniswap_pool(token_address, sync=False):
        bucket = "uni or uni-like lp"
    elif gearbox and await gearbox.is_diesel_token(token_address, sync=False):
        bucket = "gearbox"

    # this just requires contract initialization but should go behind uniswap
    elif await aave.is_wrapped_atoken_v2(token_address, sync=False):
        bucket = "wrapped atoken v2"
    elif await aave.is_wrapped_atoken_v3(token_address, sync=False):
        bucket = "wrapped atoken v3"
    elif await is_generic_amm(token_address):
        bucket = "generic amm"
    elif await mooniswap.is_mooniswap_pool(token_address, sync=False):
        bucket = "mooniswap lp"
    elif await compound.is_compound_market(token_address, sync=False):
        bucket = "compound"
    elif await _chainlink_and_band(token_address):
        bucket = "chainlink and band"
    elif chainlink and await chainlink.has_feed(token_address, sync=False):
        bucket = "chainlink feed"
    elif synthetix and await synthetix.is_synth(token_address, sync=False):
        bucket = "synthetix"
    elif await yearn.is_yearn_vault(token_address, sync=False):
        bucket = "yearn or yearn-like"
    elif curve and await curve.get_pool(token_address, sync=False):
        bucket = "curve lp"
    if debug_logs_enabled:
        await __log_bucket(token_address, bucket)
    if bucket:
        db.set_bucket(token, bucket)
    return bucket


# these require neither calls to the chain nor contract initialization, just string comparisons (pretty sure)
string_matchers = {
    "wrapped gas coin": lambda address: address
    == "0xEeeeeEeeeEeEeeEeEeEeeEEEeeeeEeeeeeeeEEeE",
    "stable usd": lambda address: address in STABLECOINS,
    "one to one": one_to_one.is_one_to_one_token,
    "wsteth": wsteth.is_wsteth,
    "creth": creth.is_creth,
    "belt lp": belt.is_belt_lp,
    "froyo": froyo.is_froyo,
    "convex": convex.is_convex_lp,
    "rkp3r": rkp3r.is_rkp3r,
}

# these just require calls
calls_only = {
    "atoken": aave.is_atoken,
    "balancer pool": balancer_multiplexer.is_balancer_pool,
    "ib token": ib.is_ib_token,
    "gelato": gelato.is_gelato_pool,
    "pendle lp": pendle.is_pendle_lp,
    "piedao lp": piedao.is_pie,
    "token set": tokensets.is_token_set,
    "ellipsis lp": ellipsis.is_eps_rewards_pool,
    "mstable feeder pool": mstablefeederpool.is_mstable_feeder_pool,
    "saddle": saddle.is_saddle_lp,
    "basketdao": basketdao.is_basketdao_index,
    "popsicle": popsicle.is_popsicle_lp,
    "reserve": reserve.is_rtoken,
}


async def _chainlink_and_band(token_address) -> bool:
    """
    Check if a token is supported by both Chainlink and Band oracles.

    This function is primarily used for historical data on the Fantom network,
    where Band was used before Chainlink became available.

    Args:
        token_address: The address of the token to check.

    Examples:
        >>> is_supported = await _chainlink_and_band("0x6B175474E89094C44Da98b954EedeAC495271d0F")
        >>> print(is_supported)
        True

    See Also:
        - :func:`y.prices.chainlink.has_feed`
    """
    return (
        chainlink
        and await chainlink.has_feed(token_address, sync=False)
        and token_address in band
    )


async def _check_bucket_helper(
    bucket: str, check: Callable[[Address], Awaitable[bool]], address: Address
) -> Tuple[str, bool]:
    """
    Helper function to check if a token belongs to a specific bucket.

    Args:
        bucket: The name of the bucket to check.
        check: A callable that checks if the token belongs to the bucket.
        address: The address of the token to check.

    Examples:
        >>> result = await _check_bucket_helper("stable usd", lambda x: x in STABLECOINS, "0x6B175474E89094C44Da98b954EedeAC495271d0F")
        >>> print(result)
        ('stable usd', True)

    See Also:
        - :func:`check_bucket`
    """
    result = await check(address, sync=False)

    # TODO: debug why we have to re-await sometimes when @optional_async_diskcache is used
    if not isinstance(result, bool):
        if not iscoroutine(result):
            raise TypeError(f"{bucket} result must be boolean. You passed {result}")
        result = await result
    if not isinstance(result, bool):
        # if not iscoroutine(result):
        raise TypeError(f"{bucket} result must be boolean. You passed {result}")
        # result = await result
    return bucket, result


async def __log_bucket(token, bucket):
    symbol = await ERC20(token, asynchronous=True).symbol
    logger._log(DEBUG, "%s %s bucket is %s", (symbol, token, bucket))


async def __log_not_bucket(token, bucket):
    symbol = await ERC20(token, asynchronous=True).symbol
    logger._log(DEBUG, "%s %s bucket is not %s", (symbol, token, bucket))



import pytest

from tests.fixtures import mainnet_only
from y.classes.common import ERC20
from y.constants import usdc
from y.prices.yearn import YearnInspiredVault

"""
These tests are written in a synchronous style using pytest. They verify the correctness of the 
YearnInspiredVault class methods when called in a synchronous context. The tests ensure that the 
YearnInspiredVault methods return expected results for given inputs, specifically focusing on 
share price and underlying asset retrieval for Yearn vaults.
"""


@mainnet_only
def test_share_price_yearn_v1():
    """Test the share price of a Yearn v1 vault."""
    yusdc_v1 = "0xa2609B2b43AC0F5EbE27deB944d2a399C201E3dA"
    block = 14_500_000
    assert YearnInspiredVault(yusdc_v1).share_price(block) == 1.111974132096754


@mainnet_only
def test_underlying_yearn_v1():
    """Test the underlying asset of a Yearn v1 vault."""
    yusdc_v1 = "0xa2609B2b43AC0F5EbE27deB944d2a399C201E3dA"
    assert YearnInspiredVault(yusdc_v1).underlying == ERC20(usdc)


@mainnet_only
def test_share_price_yearn_v2():
    """Test the share price of a Yearn v2 vault."""
    yvusdc_v2 = "0x5f18c75abdae578b483e5f43f12a39cf75b973a9"
    block = 14_500_000
    assert YearnInspiredVault(yvusdc_v2).share_price(block) == 1.096431


@mainnet_only
def test_underlying_yearn_v2():
    """Test the underlying asset of a Yearn v2 vault."""
    yvusdc_v2 = "0x5f18c75abdae578b483e5f43f12a39cf75b973a9"
    assert YearnInspiredVault(yvusdc_v2).underlying == ERC20(usdc)
